import os
from siliconcompiler.tools._common import add_require_input, add_frontend_requires, \
    get_input_files, get_tool_task, has_input_files


def setup(chip):
    '''
    Imports VHDL and converts it to verilog
    '''

    # Standard Setup
    tool = 'ghdl'
    clobber = False

    if not has_input_files(chip, 'input', 'rtl', 'vhdl'):
        return "no files in [input,rtl,vhdl]"

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'exe', 'ghdl')
    chip.set('tool', tool, 'vswitch', '--version')
    chip.set('tool', tool, 'version', '>=4.0.0-dev', clobber=clobber)

    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index, clobber=clobber)
    chip.set('tool', tool, 'task', task, 'option', '',
             step=step, index=index, clobber=clobber)
    chip.set('tool', tool, 'task', task, 'stdout', 'destination', 'output',
             step=step, index=index)
    chip.set('tool', tool, 'task', task, 'stdout', 'suffix', 'v',
             step=step, index=index)

    # Schema requirements
    add_require_input(chip, 'input', 'rtl', 'vhdl')
    add_frontend_requires(chip, ['define'])

    design = chip.top()

    chip.set('tool', tool, 'task', task, 'output', f'{design}.v', step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'extraopts', 'extra options to pass to ghdl',
             field='help')
    if chip.get('tool', tool, 'task', task, 'var', 'extraopts', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['tool', tool, 'task', task, 'var', 'extraopts']),
                 step=step, index=index)


################################
#  Custom runtime options
################################
def runtime_options(chip):

    ''' Custom runtime options, returnst list of command line options.
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    options = []

    # Synthesize inputs and output Verilog netlist
    options.append('--synth')
    options.append('--std=08')
    options.append('--out=verilog')
    options.append('--no-formal')

    # currently only -fsynopsys and --latches supported
    valid_extraopts = ['-fsynopsys', '--latches']

    extra_opts = chip.get('tool', 'ghdl', 'task', task, 'var', 'extraopts', step=step, index=index)
    for opt in extra_opts:
        if opt in valid_extraopts:
            options.append(opt)
        else:
            chip.error('Unsupported option ' + opt)

    # Add defines
    for define in chip.get('option', 'define'):
        options.append(f'-g{define}')

    # Add sources
    for value in get_input_files(chip, 'input', 'rtl', 'vhdl'):
        options.append(value)

    # Set top module
    options.append('-e')

    options.append(chip.top(step=step, index=index))

    return options



from siliconcompiler.tools.builtin import _common
from siliconcompiler.tools.builtin.builtin import set_io_files
from siliconcompiler import flowgraph


def setup(chip):
    '''
    A no-operation that passes inputs to outputs.
    '''

    set_io_files(chip)


def _select_inputs(chip, step, index):
    return _common._select_inputs(chip, step, index)


def _gather_outputs(chip, step, index):
    '''Return set of filenames that are guaranteed to be in outputs
    directory after a successful run of step/index.'''

    flow = chip.get('option', 'flow')

    in_nodes = chip.get('flowgraph', flow, step, index, 'input')
    in_task_outputs = [flowgraph._gather_outputs(chip, *node) for node in in_nodes]

    if len(in_task_outputs) > 0:
        return in_task_outputs[0].union(*in_task_outputs[1:])

    return []


def run(chip):
    return _common.run(chip)


def post_process(chip):
    _common.post_process(chip)



'''
This tool is used to execute the output of a previous step.
For example, if the flow contains a compile step which generates the
next executable needed in the flow.
'''

from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'exe', ":exe:", clobber=False)
    chip.set('tool', tool, 'task', task, 'option', [], step=step, index=index, clobber=False)



import os
import shutil

from siliconcompiler.tools.vpr import vpr
from siliconcompiler.tools._common import get_tool_task


def setup(chip, clobber=True):
    '''
    Perform automated place and route with VPR
    '''

    tool = 'vpr'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    vpr.setup_tool(chip, clobber=clobber)

    # Match VPR's default max router iterations value so that
    # the parameter is traceable by SC through the flow.
    chip.set('tool', tool, 'task', task, 'var', 'max_router_iterations', "50",
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'var', 'max_router_iterations',
             'set maximum number of routing iterations',
             field='help')

    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['tool', tool, 'task', task, 'var', 'max_router_iterations']),
             step=step, index=index)

    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index, clobber=clobber)

    # TO-DO: PRIOROTIZE the post-routing packing results?
    design = chip.top()
    chip.set('tool', tool, 'task', task, 'input', design + '.blif', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'input', design + '.net', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'input', design + '.place', step=step, index=index)

    chip.set('tool', tool, 'task', task, 'output', design + '.route', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', design + '.blif', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', design + '.net', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', design + '.place', step=step, index=index)


def runtime_options(chip):
    '''Command line options to vpr for the route step
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    design = chip.top()

    options = vpr.runtime_options(chip)

    blif = f"inputs/{design}.blif"
    options.append(blif)

    options.append('--route')
    # To run only the routing step we need to pass in the placement files
    options.append(f'--net_file inputs/{design}.net')
    options.append(f'--place_file inputs/{design}.place')

    enable_images = chip.get('tool', tool, 'task', task, 'var', 'enable_images',
                             step=step, index=index)[0]

    route_iterations = chip.get('tool', tool, 'task', task, 'var', 'max_router_iterations',
                                step=step, index=index)

    if (len(route_iterations) > 0):
        options.append(f'--max_router_iterations {route_iterations[0]}')

    if enable_images == 'true':
        design = chip.top()

        graphics_commands = vpr.get_common_graphics(chip)

        # set_draw_block_text 0 hides the label for various blocks in the design
        # set_draw_block_outlines 0 removes the outline/boundary for various blocks in the design
        # set_routing_util 1 displays the routing utilization as a heat map
        # set_routing_util 4 displays the routing utilization as a heat map over placed blocks
        # Refer: https://github.com/verilog-to-routing/vtr-verilog-to-routing/blob/master/
        # vpr/src/draw/draw_types.h#L89
        # save_graphics saves the block diagram as a png/svg/pdf
        # Refer:
        # https://docs.verilogtorouting.org/en/latest/vpr/command_line_usage/#graphics-options
        graphics_commands.append("set_draw_block_text 0; " +
                                 "set_draw_block_outlines 0; " +
                                 "set_routing_util 1; " +
                                 "save_graphics "
                                 f"reports/{design}_route_utilization_with_placement.png;")
        graphics_commands.append("set_draw_block_text 0; " +
                                 "set_draw_block_outlines 0; " +
                                 "set_routing_util 4; " +
                                 "save_graphics "
                                 f"reports/{design}_route_utilization.png;")

        graphics_command_str = " ".join(graphics_commands)

        options.append("--graphics_commands")
        options.append(f"\"{graphics_command_str}\"")

    return options


################################
# Post_process (post executable)
################################


def post_process(chip):
    ''' Tool specific function to run after step execution
    '''
    vpr.vpr_post_process(chip)

    design = chip.top()
    # Forward all of the prior step inputs forward for bitstream generation
    shutil.copy2(f'inputs/{design}.blif', 'outputs')
    shutil.copy2(f'inputs/{design}.net', 'outputs')
    shutil.copy2(f'inputs/{design}.place', 'outputs')



import os
import shutil
from siliconcompiler.tools._common import \
    add_frontend_requires, add_require_input, get_frontend_options, get_input_files, \
    get_tool_task, has_input_files


def setup(chip):
    '''
    Performs high level synthesis to generate a verilog output
    '''

    if not has_input_files(chip, 'input', 'hll', 'c'):
        return "no files in [input,hll,c]"

    tool = 'bambu'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    # Standard Setup
    refdir = 'tools/' + tool
    chip.set('tool', tool, 'exe', 'bambu')
    chip.set('tool', tool, 'vswitch', '--version')
    chip.set('tool', tool, 'version', '>=2024.03', clobber=False)

    chip.set('tool', tool, 'task', task, 'refdir', refdir,
             step=step, index=index,
             package='siliconcompiler', clobber=False)
    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index, clobber=False)

    # Input/Output requirements
    chip.add('tool', tool, 'task', task, 'output', chip.top() + '.v', step=step, index=index)

    # Schema requirements
    add_require_input(chip, 'input', 'hll', 'c')
    add_frontend_requires(chip, ['idir', 'define'])


################################
#  Custom runtime options
################################
def runtime_options(chip):
    cmdlist = []

    cmdlist.append('--memory-allocation-policy=NO_BRAM')

    opts = get_frontend_options(chip, ['idir', 'define'])

    for value in opts['idir']:
        cmdlist.append('-I' + value)
    for value in opts['define']:
        cmdlist.append('-D' + value)
    for value in get_input_files(chip, 'input', 'hll', 'c'):
        cmdlist.append(value)

    cmdlist.append('--memory-allocation-policy=NO_BRAM')

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    cmdlist.append(f'--top-fname={chip.top(step, index)}')

    return cmdlist


################################
# Post_process (post executable)
################################
def post_process(chip):
    ''' Tool specific function to run after step execution
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    shutil.copy2(f'{chip.top(step, index)}.v', os.path.join('outputs', f'{chip.top()}.v'))



import os
import siliconcompiler
from siliconcompiler.targets import skywater130_demo


def setup(chip):
    '''
    "Self-test" target which builds a small 8-bit counter design as an ASIC,
    targeting the Skywater130 PDK.

    This target is intended for testing purposes only,
    to verify that SiliconCompiler is installed and configured correctly.
    '''

    # Set design name
    design = 'heartbeat'
    chip.set('design', design)

    # Load the Sky130 PDK/standard cell library target.
    chip.use(skywater130_demo)

    # Set quiet flag
    chip.set('option', 'quiet', True, clobber=False)

    # Set die area and clock constraint.
    chip.set('constraint', 'outline', [(0, 0), (50, 50)], clobber=False)
    chip.set('constraint', 'corearea', [(5, 5), (45, 45)], clobber=False)
    chip.clock('clk', period=10)

    # Add source files.
    chip.input(os.path.join('data', f'{design}.v'), package='siliconcompiler')


#########################
if __name__ == "__main__":
    target = siliconcompiler.Chip('<target>')
    setup(target)
    target.write_manifest('asic_demo.json')



import siliconcompiler

from siliconcompiler.flows._common import setup_multiple_frontends
from siliconcompiler.flows._common import _make_docs

from siliconcompiler.tools.yosys import syn_asic
from siliconcompiler.tools.openroad import floorplan
from siliconcompiler.tools.openroad import physyn
from siliconcompiler.tools.openroad import place
from siliconcompiler.tools.openroad import cts
from siliconcompiler.tools.openroad import route
from siliconcompiler.tools.openroad import dfm
from siliconcompiler.tools.openroad import export as openroad_export
from siliconcompiler.tools.klayout import export as klayout_export

from siliconcompiler.tools.builtin import minimum


############################################################################
# DOCS
############################################################################
def make_docs(chip):
    n = 3
    _make_docs(chip)
    return setup(syn_np=n, floorplan_np=n, physyn_np=n, place_np=n, cts_np=n, route_np=n)


###########################################################################
# Flowgraph Setup
############################################################################
def setup(flowname='asicflow',
          syn_np=1,
          floorplan_np=1,
          physyn_np=1,
          place_np=1,
          cts_np=1,
          route_np=1):
    '''
    A configurable ASIC compilation flow.

    The 'asicflow' includes the stages below. The steps syn, floorplan,
    physyn, place, cts, route, and dfm have minimization associated
    with them. To view the flowgraph, see the .png file.

    * **import**: Sources are collected and packaged for compilation
    * **syn**: Translates RTL to netlist using Yosys
    * **floorplan**: Floorplanning
    * **physyn**: Physical Synthesis
    * **place**: Global and detailed placement
    * **cts**: Clock tree synthesis
    * **route**: Global and detailed routing
    * **dfm**: Metal fill, atenna fixes and any other post routing steps
    * **export**: Export design from APR tool and merge with library GDS
    * **sta**: Static timing analysis (signoff)
    * **lvs**: Layout versus schematic check (signoff)
    * **drc**: Design rule check (signoff)

    The syn, physyn, place, cts, route steps supports per process
    options that can be set up by setting '<step>_np'
    arg to a value > 1, as detailed below:

    * syn_np : Number of parallel synthesis jobs to launch
    * floorplan_np : Number of parallel floorplan jobs to launch
    * physyn_np : Number of parallel physical synthesis jobs to launch
    * place_np : Number of parallel place jobs to launch
    * cts_np : Number of parallel clock tree synthesis jobs to launch
    * route_np : Number of parallel routing jobs to launch
    '''

    flow = siliconcompiler.Flow(flowname)

    # Linear flow, up until branch to run parallel verification steps.
    longpipe = ['syn',
                'synmin',
                'floorplan',
                'floorplanmin',
                'place',
                'placemin',
                'cts',
                'ctsmin',
                'route',
                'routemin',
                'dfm']

    # step --> task
    tasks = {
        'syn': syn_asic,
        'synmin': minimum,
        'floorplan': floorplan,
        'floorplanmin': minimum,
        'physyn': physyn,
        'physynmin': minimum,
        'place': place,
        'placemin': minimum,
        'cts': cts,
        'ctsmin': minimum,
        'route': route,
        'routemin': minimum,
        'dfm': dfm
    }

    np = {
        "syn": syn_np,
        "floorplan": floorplan_np,
        "physyn": physyn_np,
        "place": place_np,
        "cts": cts_np,
        "route": route_np
    }

    prevstep = None
    # Remove built in steps where appropriate
    flowpipe = []
    for step in longpipe:
        task = tasks[step]
        if task == minimum:
            if prevstep in np and np[prevstep] > 1:
                flowpipe.append(step)
        else:
            flowpipe.append(step)
        prevstep = step

    flowtasks = []
    for step in flowpipe:
        flowtasks.append((step, tasks[step]))

    # Programmatically build linear portion of flowgraph and fanin/fanout args
    prevstep = setup_multiple_frontends(flow)
    for step, task in flowtasks:
        fanout = 1
        if step in np:
            fanout = np[step]
        # create nodes
        for index in range(fanout):
            # nodes
            flow.node(flowname, step, task, index=index)

            # edges
            if task == minimum:
                fanin = 1
                if prevstep in np:
                    fanin = np[prevstep]
                for i in range(fanin):
                    flow.edge(flowname, prevstep, step, tail_index=i)
            elif prevstep:
                flow.edge(flowname, prevstep, step, head_index=index)

            # metrics
            goal_metrics = ()
            weight_metrics = ()
            if task in (syn_asic, ):
                goal_metrics = ('errors',)
                weight_metrics = ()
            elif task in (floorplan, physyn, place, cts, route, dfm):
                goal_metrics = ('errors', 'setupwns', 'setuptns')
                weight_metrics = ('cellarea', 'peakpower', 'leakagepower')

            for metric in goal_metrics:
                flow.set('flowgraph', flowname, step, str(index), 'goal', metric, 0)
            for metric in weight_metrics:
                flow.set('flowgraph', flowname, step, str(index), 'weight', metric, 1.0)
        prevstep = step

    # add write information steps
    flow.node(flowname, 'write_gds', klayout_export)
    flow.edge(flowname, prevstep, 'write_gds')
    flow.node(flowname, 'write_data', openroad_export)
    flow.edge(flowname, prevstep, 'write_data')

    return flow


##################################################
if __name__ == "__main__":
    chip = siliconcompiler.Chip('design')
    chip.set('input', 'constraint', 'sdc', 'test')
    flow = make_docs(chip)
    chip.use(flow)
    chip.write_flowgraph(f"{flow.top()}.png", flow=flow.top())



import json


def load_json_constraints(json_constraints_file):

    json_generic_constraints = {}
    with (open(json_constraints_file, "r")) as json_constraints_data:
        json_generic_constraints = json.loads(json_constraints_data.read())

    return json_generic_constraints


def load_constraints_map(constraints_map_file):

    constraints_map = {}
    with (open(constraints_map_file, "r")) as constraints_map_data:
        constraints_map = json.loads(constraints_map_data.read())

    return constraints_map


def map_constraints(chip,
                    json_generic_constraints,
                    constraints_map):

    design_constraints = {}
    errors = 0

    # If no constraints map is provided pass the constraints directly
    if not constraints_map:

        for design_pin in json_generic_constraints:
            design_constraints[design_pin] = json_generic_constraints[design_pin]['pin']

    # Otherwise use the constraints map to remap the constraints to the correct
    # internal FPGA core pin:
    else:
        for design_pin in json_generic_constraints:

            # VPR has a quirk that it prepends "out:" to outputs to differentiate
            # the pin from any block name that may have inherited its instance name
            # from an output.  Compensate for that quirk here
            if (json_generic_constraints[design_pin]['direction'] == "output"):
                named_design_pin = f'out:{design_pin}'
            else:
                named_design_pin = design_pin

            design_pin_assignment = json_generic_constraints[design_pin]['pin']

            if (design_pin_assignment in constraints_map):
                # Convert the dictionary entries in the constraints map into a tuple:
                design_pin_constraint_assignment = (
                    constraints_map[design_pin_assignment]['x'],
                    constraints_map[design_pin_assignment]['y'],
                    constraints_map[design_pin_assignment]['subtile'])

                design_constraints[named_design_pin] = design_pin_constraint_assignment

            else:
                chip.logger.error(f'Cannot map to pin {design_pin_assignment}')
                errors += 1

    return design_constraints, errors



from siliconcompiler.tools.klayout import show as klayout_show
from siliconcompiler.tools.klayout import screenshot as klayout_screenshot
from siliconcompiler.tools.openroad import show as openroad_show
from siliconcompiler.tools.openroad import screenshot as openroad_screenshot
from siliconcompiler.tools.vpr import show as vpr_show
from siliconcompiler.tools.vpr import screenshot as vpr_screenshot
from siliconcompiler.tools.yosys import screenshot as yosys_screenshot


def setup(chip):
    chip.register_showtool('gds', klayout_show)
    chip.register_showtool('gds', klayout_screenshot)
    chip.register_showtool('oas', klayout_show)
    chip.register_showtool('oas', klayout_screenshot)
    chip.register_showtool('lef', klayout_show)
    chip.register_showtool('lef', klayout_screenshot)

    chip.register_showtool('odb', openroad_show)
    chip.register_showtool('odb', openroad_screenshot)
    chip.register_showtool('def', openroad_show)
    chip.register_showtool('def', openroad_screenshot)

    chip.register_showtool('route', vpr_show)
    chip.register_showtool('route', vpr_screenshot)
    chip.register_showtool('place', vpr_show)
    chip.register_showtool('place', vpr_screenshot)

    chip.register_showtool('v', yosys_screenshot)
    chip.register_showtool('vg', yosys_screenshot)



from siliconcompiler.tools.builtin import _common
from siliconcompiler.tools.builtin import minimum
from siliconcompiler.tools.builtin.builtin import set_io_files


def setup(chip):
    '''
    Selects the task with the maximum metric score from a list of inputs.

    Sequence of operation:

    1. Check list of input tasks to see if all metrics meets goals
    2. Check list of input tasks to find global min/max for each metric
    3. Select MAX value if all metrics are met.
    4. Normalize the min value as sel = (val - MIN) / (MAX - MIN)
    5. Return normalized value and task name

    Meeting metric goals takes precedence over compute metric scores.
    Only goals with values set and metrics with weights set are considered
    in the calculation.
    '''

    set_io_files(chip)


def _select_inputs(chip, step, index):
    inputs = _common._select_inputs(chip, step, index)

    score, sel_inputs = _common._minmax(chip, *inputs, op='maximum')

    if sel_inputs:
        chip.logger.info(f"Selected '{sel_inputs[0]}{sel_inputs[1]}' with score {score:.3f}")

    return sel_inputs


def _gather_outputs(chip, step, index):
    return minimum._gather_outputs(chip, step, index)


def run(chip):
    return _common.run(chip)


def post_process(chip):
    _common.post_process(chip)



import os
import shutil
from siliconcompiler.tools.genfasm import genfasm
from siliconcompiler.tools.vpr import vpr
from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    '''
    Generates a bitstream
    '''
    genfasm.setup(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'regex', 'warnings', "^Warning",
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'regex', 'errors', "^Error",
             step=step, index=index, clobber=False)

    design = chip.top()

    chip.set('tool', tool, 'task', task, 'input', design + '.route', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'input', design + '.blif', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'input', design + '.net', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'input', design + '.place', step=step, index=index)

    chip.add('tool', tool, 'task', task, 'output', design + '.fasm', step=step, index=index)


def runtime_options(chip):
    options = vpr.runtime_options(chip)

    design = chip.top()

    blif = f"inputs/{design}.blif"
    options.append(blif)

    options.append(f'--net_file inputs/{design}.net')
    options.append(f'--place_file inputs/{design}.place')
    options.append(f'--route_file inputs/{design}.route')

    return options

################################
# Post_process (post executable)
################################


def post_process(chip):
    ''' Tool specific function to run after step execution
    '''
    vpr.vpr_post_process(chip)

    design = chip.top()
    shutil.move(f'{design}.fasm', 'outputs')



import math
import re

SI_UNITS = (
    ('y', -24),
    ('z', -21),
    ('a', -18),
    ('f', -15),
    ('p', -12),
    ('n', -9),
    ('u', -6),
    ('m', -3),
    ('', 0),
    ('k', 3),
    ('M', 6),
    ('G', 9),
    ('T', 12),
    ('P', 15),
    ('E', 18),
    ('Z', 21),
    ('Y', 24)
)

BINARY_UNITS = (
    ('', 0),
    ('k', 10),
    ('M', 20),
    ('G', 30),
    ('T', 40),
    ('P', 50),
    ('E', 60),
    ('Z', 70)
)

SI_TYPES = (
    's',
    'Hz',
    'F',
    'm',
    'A',
    'V',
    'W',
    'ohm',
    'C',
)


BINARY_TYPES = (
    'B',
    'b'
)


def convert(value, from_unit=None, to_unit=None):
    '''
    Convert a value to from one SI power to another SI power

    Args:
        value (float): value to convert
        from_unit (str): unit of the value, default is None and assumes no magnitude
        to_unit (str): unit of the return, default is None and assumes no magnitude

    Returns:
        float: scaled value
    '''
    value = float(value)

    power = get_si_power(to_unit)

    from_scale = _get_scale(from_unit) ** power
    to_scale = _get_scale(to_unit) ** power
    scale = from_scale / to_scale
    if scale > 1:
        scale = round(scale)
    elif scale < 1:
        scale = 1.0 / round(1.0 / scale)
    else:
        scale = round(scale)

    return value * scale


def _get_scale(unit):
    if not unit:
        unit = ''
    unit_prefix = get_si_prefix(unit)
    for prefix, scale in SI_UNITS:
        if prefix == unit_prefix:
            return 10**scale

    return 1


def get_si_prefix(unit):
    '''
    Get the SI prefix of the specific unit.

    For example:
    get_si_prefix('um') -> 'u'
    '''
    if not unit:
        return ''

    for si_type in list(SI_TYPES) + ['']:
        re_find = fr'^(.*){si_type}(\^[0-9]+)?$'
        matches = re.findall(re_find, unit, re.IGNORECASE)
        if matches:
            return matches[0][0]

    return ''


def get_si_power(unit):
    '''
    Get the SI power of the specific unit.
    This is mainly needed for area units.

    For example:
    get_si_prefix('um') -> 1
    get_si_prefix('um^2') -> 2
    '''
    if not unit:
        return 1

    for si_type in list(SI_TYPES) + ['']:
        re_find = fr'^.*{si_type}\^([0-9]+)$'
        matches = re.findall(re_find, unit, re.IGNORECASE)
        if matches:
            return int(matches[0][-1])

    return 1


def is_base_si_unit(unit):
    '''
    Check if a unit has no magnitude
    '''
    return unit in SI_TYPES


def is_base_si_unit_power(unit):
    '''
    Check if a unit has a power associated with it
    '''
    return get_si_power(unit) > 1


def is_base_binary_unit(unit):
    '''
    Check if a unit is binary
    '''
    return unit in BINARY_TYPES


def format_si(value, unit, margin=3, digits=3):
    '''
    Format a number as an SI number. Returns a string.

    Args:
        value (float): value to convert
        unit (str): unit of the value
        margin (int): number of extra digits to ensure are preserved
            when picking the right magnitude
        digits (int): number of digits to print after .
    '''
    scaled_value, prefix = scale_si(value, unit, margin=margin, digits=digits)
    # need to do this in case float shortens scaled_value
    return f'{scaled_value:.{digits}f}{prefix}'


def scale_si(value, unit, margin=3, digits=3):
    '''
    Format a number as an SI number. Returns a float.

    Args:
        value (float): value to convert
        unit (str): unit of the value
        margin (int): number of extra digits to ensure are preserved
            when picking the right magnitude
        digits (int): number of digits to print after .
    '''
    if unit and is_base_si_unit(unit):
        value = float(value)
        log_value = math.log10(value) - margin

        if digits < 0:
            # Default to 0
            digits = 0

        for prefix, scale in SI_UNITS:
            if log_value <= scale:
                value /= 10**scale
                return (float(f'{value:.{digits}f}'), prefix)

        return (float(f'{value:.{digits}f}'), '')

    return (float(f'{value:.{digits}f}'), '')


def format_binary(value, unit, digits=3):
    '''
    Format a number as a binary number. Returns a string.

    Args:
        value (float): value to convert
        unit (str): unit of the value
        digits (int): number of digits to print after .
    '''
    scaled_value, prefix = scale_binary(value, unit, digits=digits)
    # need to do this in case float shortens scaled_value
    return f'{scaled_value:.{digits}f}{prefix}'


def scale_binary(value, unit, digits=3):
    '''
    Format a number as a binary number. Returns a float.

    Args:
        value (float): value to convert
        unit (str): unit of the value
        digits (int): number of digits to print after .
    '''
    value = float(value)

    fvalue = (int(value), '')
    if is_base_binary_unit(unit):
        for prefix, scale in BINARY_UNITS:
            new_value = value / 2**scale

            if new_value > 1:
                fvalue = (float(f'{new_value:.{digits}f}'), prefix)
                continue

            return fvalue

    return (float(f'{value:.{digits}f}'), '')


def format_time(value):
    '''
    Format a number as time.
    Prints as hh:mm:ss.ms (hours:minutes:seconds.milliseconds)

    Args:
        value (float): number of seconds to convert
    '''
    # Report as hh:mm::ss.ms
    value, milliseconds = divmod(value, 1)
    hours, value = divmod(value, 3600)
    minutes, seconds = divmod(value, 60)
    milliseconds *= 1000
    ftime = ''
    if hours > 0:
        ftime += f'{int(hours)}:'
    if hours > 0 or minutes > 0:
        ftime += f'{int(minutes):02}:'
    ftime += f'{int(seconds):02}.'
    ftime += f'{int(milliseconds):03}'
    return ftime



import streamlit

from siliconcompiler.report.dashboard import components
from siliconcompiler.report.dashboard import layouts
from siliconcompiler.report.dashboard import state
from siliconcompiler.report.dashboard import utils

import streamlit_autorefresh


if __name__ == "__main__":
    # opened by running command in siliconcompiler/apps/sc_dashboard.py
    state.init()

    components.setup_page()
    state.setup()

    layout = layouts.get_layout(state.get_key(state.APP_LAYOUT))
    layout()

    reload = False
    if state.get_key(state.SELECTED_JOB) == 'default':
        reload = state.set_key(state.IS_RUNNING, utils.is_running(state.get_chip()))

    if state.get_key(state.IS_RUNNING):
        update_interval = state.get_key(state.APP_RUNNING_REFRESH)
    else:
        update_interval = state.get_key(state.APP_STOPPED_REFRESH)

    streamlit_autorefresh.st_autorefresh(interval=update_interval)

    state.debug_print_state()

    if reload or state.update_manifest() or state.get_key(state.APP_RERUN):
        state.set_key(state.APP_RERUN, None)
        streamlit.rerun()



# KLayout script to export an OpenROAD marker DB from a DRC db
#
# Based on:
# https://github.com/The-OpenROAD-Project/OpenROAD-flow-scripts/blob/master/flow/util/convertDrc.py

import pya
import glob
import json
import os
import sys


def convert_drc(view, path):
    rdb_id = view.create_rdb(os.path.basename(path))
    rdb = view.rdb(rdb_id)
    print(f"[INFO] reading {path}")
    rdb.load(path)

    source = os.path.abspath(path)

    ordb = {
        "source": source,
        "description": "KLayout DRC conversion",
        "category": {}
    }

    for category in rdb.each_category():
        if category.num_items() == 0:
            # ignore categories with no data
            continue

        ordb_category = {
            "description": category.description,
            "source": source,
            "violations": []
        }
        ordb["category"][category.name()] = ordb_category

        for item in rdb.each_item_per_category(category.rdb_id()):
            violation = {
                "visited": item.is_visited(),
                "visible": True,
                "waived": "waived" in item.tags_str
            }

            ordb_category["violations"].append(violation)

            shapes = []
            violation["shape"] = shapes

            text = []

            for value in item.each_value():
                if value.is_box():
                    shapes.append({
                        "type": "box",
                        "points": [{
                            "x": value.box().left,
                            "y": value.box().bottom
                        }, {
                            "x": value.box().right,
                            "y": value.box().top
                        }]
                    })
                elif value.is_edge():
                    shapes.append({
                        "type": "line",
                        "points": [{
                            "x": value.edge().p1.x,
                            "y": value.edge().p1.y
                        }, {
                            "x": value.edge().p2.x,
                            "y": value.edge().p2.y
                        }]
                    })
                elif value.is_edge_pair():
                    edge1 = value.edge_pair().first
                    edge2 = value.edge_pair().second

                    shapes.append({
                        "type": "line",
                        "points": [{
                            "x": edge1.p1.x,
                            "y": edge1.p1.y
                        }, {
                            "x": edge1.p2.x,
                            "y": edge1.p2.y
                        }]
                    })
                    shapes.append({
                        "type": "line",
                        "points": [{
                            "x": edge2.p1.x,
                            "y": edge2.p1.y
                        }, {
                            "x": edge2.p2.x,
                            "y": edge2.p2.y
                        }]
                    })
                elif value.is_polygon():
                    points = []
                    for edge in value.polygon().each_edge():
                        points.append({
                            "x": edge.p1.x,
                            "y": edge.p1.y
                        })
                    points.append({
                        "x": edge.p2.x,
                        "y": edge.p2.y
                    })
                    shapes.append({
                        "type": "polygon",
                        "points": points
                    })
                elif value.is_path():
                    points = []
                    for edge in value.path().polygon().each_edge():
                        points.append({
                            "x": edge.p1.x,
                            "y": edge.p1.y
                        })
                    points.append({
                        "x": edge.p2.x,
                        "y": edge.p2.y
                    })
                    shapes.append({
                        "type": "polygon",
                        "points": points
                    })
                elif value.is_text():
                    text.append(value.text())
                elif value.is_string():
                    text.append(value.string())
                else:
                    print("[WARN] Unknown violation shape:", value)

            comment = ""
            if hasattr(item, 'comment'):
                comment = item.comment
            if text:
                if comment:
                    comment += ": "
                comment += ", ".join(text)

            if comment:
                violation["comment"] = comment

    return ordb


def main():
    # SC_ROOT provided by CLI
    sys.path.append(SC_ROOT)  # noqa: F821

    from tools.klayout.klayout_utils import get_schema

    schema = get_schema(manifest='sc_manifest.json')

    design = schema.get('design')

    app = pya.Application.instance()
    win = app.main_window()

    # Create a dummy view to use for loading
    cell_view = win.create_layout(0)
    layout_view = cell_view.view()

    ordb = {}
    for file in glob.glob(f'inputs/{design}*.lyrdb') + glob.glob('inputs/{design}*.ascii'):
        name = os.path.basename(file)

        ordb[name] = convert_drc(layout_view, file)

    with open(f"outputs/{design}.json", "w") as outfile:
        json.dump(
            ordb,
            outfile,
            indent=2)


if __name__ == '__main__':
    main()



import argparse
import json
import os
import streamlit
import streamlit_javascript
import fasteners

from siliconcompiler import Chip


DISPLAY_FLOWGRAPH = "show_flowgraph"
SELECTED_JOB = "selected_job"
SELECTED_NODE = "selected_node"
# This is needed until the graph supports setting a selected node
SELECTED_FLOWGRAPH_NODE = "selected_flowgraph_node"
SELECTED_SELECTOR_NODE = "selected_selector_node"
NODE_SOURCE = "node_source"
SELECTED_FILE = "selected_file"
SELECTED_FILE_PAGE = "selected_file_page"
LOADED_CHIPS = "loaded_chips"
UI_WIDTH = "ui_width"
MANIFEST_FILE = "manifest_file"
MANIFEST_LOCK = "manifest_lock"
MANIFEST_TIME = "manifest_time"
IS_RUNNING = "is_flow_running"
GRAPH_JOBS = "graph_jobs"
APP_LAYOUT = "app_layout"
APP_RERUN = "app_rerun"
APP_RUNNING_REFRESH = "app_running_refresh"
APP_STOPPED_REFRESH = "app_stopped_refresh"
MAX_DICT_ITEMS_TO_SHOW = "max_dict_items"
MAX_FILE_LINES_TO_SHOW = "max_file_lines"
SELECT_TAB = "select_tab"
TAB_INDEX = "tab-index"
TAB_STATE = "tab-state"

_DEBUG = False
DEVELOPER = False


def _add_default(key, value):
    if key not in streamlit.session_state:
        streamlit.session_state[key] = value


def update_manifest():
    file_time = os.stat(get_key(MANIFEST_FILE)).st_mtime

    if get_key(MANIFEST_TIME) != file_time:
        chip = Chip(design='')

        with get_key(MANIFEST_LOCK):
            chip.read_manifest(get_key(MANIFEST_FILE))
        set_key(MANIFEST_TIME, file_time)
        debug_print("Read manifest", get_key(MANIFEST_FILE))

        add_chip("default", chip)

        for history in chip.getkeys('history'):
            history_chip = Chip(design='')
            history_chip.schema.cfg = chip.getdict('history', history)
            history_chip.set('design', chip.design)
            add_chip(history, history_chip)

        return True
    return False


def init():
    _add_default(DISPLAY_FLOWGRAPH, True)
    _add_default(SELECTED_JOB, None)
    _add_default(SELECTED_NODE, None)
    _add_default(SELECTED_FLOWGRAPH_NODE, None)
    _add_default(SELECTED_SELECTOR_NODE, None)
    _add_default(NODE_SOURCE, None)
    _add_default(SELECTED_FILE, None)
    _add_default(SELECTED_FILE_PAGE, None)
    _add_default(LOADED_CHIPS, {})
    _add_default(MANIFEST_FILE, None)
    _add_default(MANIFEST_LOCK, None)
    _add_default(MANIFEST_TIME, None)
    _add_default(IS_RUNNING, False)
    _add_default(GRAPH_JOBS, None)
    _add_default(UI_WIDTH, None)
    _add_default(APP_LAYOUT, "vertical_flowgraph_sac_tabs")
    _add_default(APP_RERUN, None)
    _add_default(APP_RUNNING_REFRESH, 2 * 1000)
    _add_default(APP_STOPPED_REFRESH, 30 * 1000)
    _add_default(MAX_DICT_ITEMS_TO_SHOW, 100)
    _add_default(MAX_FILE_LINES_TO_SHOW, 100)
    _add_default(SELECT_TAB, None)
    _add_default(TAB_INDEX, 0)

    parser = argparse.ArgumentParser('dashboard')
    parser.add_argument('cfg', nargs='?')
    args = parser.parse_args()

    if not args.cfg:
        raise ValueError('configuration not provided')

    if not get_key(LOADED_CHIPS):
        # First time through

        with open(args.cfg, 'r') as f:
            config = json.load(f)

        set_key(MANIFEST_FILE, config["manifest"])
        set_key(MANIFEST_LOCK, fasteners.InterProcessLock(config["lock"]))

        update_manifest()
        chip = get_chip("default")
        for graph_info in config['graph_chips']:
            file_path = graph_info['path']
            graph_chip = Chip(design='')
            graph_chip.read_manifest(file_path)
            graph_chip.unset('arg', 'step')
            graph_chip.unset('arg', 'index')

            if graph_info['cwd']:
                graph_chip.cwd = graph_info['cwd']

            add_chip(os.path.basename(file_path), graph_chip)

        chip_step = chip.get('arg', 'step')
        chip_index = chip.get('arg', 'index')

        if chip_step and chip_index:
            set_key(SELECTED_NODE, f'{chip_step}{chip_index}')

    chip = get_chip("default")
    chip.unset('arg', 'step')
    chip.unset('arg', 'index')

    if not get_key(SELECTED_JOB):
        set_key(SELECTED_JOB, "default")


def setup():
    with streamlit.empty():
        # get width
        set_key(UI_WIDTH, streamlit_javascript.st_javascript("window.innerWidth"))
        # replace with a empty container to avoid adding a gap at the top
        streamlit.empty()

    set_key(NODE_SOURCE, None)


def get_chip(job=None):
    if not job:
        job = get_key(SELECTED_JOB)
    return get_key(LOADED_CHIPS)[job]


def add_chip(name, chip):
    streamlit.session_state[LOADED_CHIPS][name] = chip


def get_chips():
    chips = list(get_key(LOADED_CHIPS).keys())
    chips.remove('default')
    chips.insert(0, 'default')
    return chips


def get_selected_node():
    if get_key(NODE_SOURCE) == "flowgraph":
        return get_key(SELECTED_FLOWGRAPH_NODE)

    if get_key(NODE_SOURCE) == "selector":
        return get_key(SELECTED_SELECTOR_NODE)

    return get_key(SELECTED_NODE)


def get_key(key):
    return streamlit.session_state[key]


def set_key(key, value):
    changed = value != streamlit.session_state[key]
    if changed:
        debug_print("set_key()", key, "changed", streamlit.session_state[key], "->", value)
        streamlit.session_state[key] = value
        return True
    return False


def del_key(key):
    debug_print("del_key()", key)
    if key in streamlit.session_state:
        del streamlit.session_state[key]


def compute_component_size(minimum, requested_px):
    ui_width = get_key(UI_WIDTH)

    if ui_width > 0:
        return min(requested_px / ui_width, minimum)

    return minimum


def debug_print(*args):
    if not _DEBUG:
        return

    print(*args)


def debug_print_state():
    if not _DEBUG:
        return

    for n, key in enumerate(sorted(streamlit.session_state.keys())):
        value = streamlit.session_state[key]
        print("state", n, key, type(value), value)
    print()



from siliconcompiler.tools.yosys.yosys import setup as tool_setup
import os
import siliconcompiler.tools.yosys.prepareLib as prepareLib
from siliconcompiler.tools._common.asic import get_libraries
from siliconcompiler.tools._common import get_tool_task
from siliconcompiler.targets import asap7_demo


def make_docs(chip):
    chip.use(asap7_demo)


def setup(chip):
    '''
    Generate a screenshot of the design
    '''

    # Generic tool setup.
    tool_setup(chip)

    # ASIC-specific setup.
    # setup_asic(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    chip.set('tool', tool, 'task', task, 'input', [], step=step, index=index)
    chip.set('tool', tool, 'task', task, 'script', 'sc_screenshot.tcl',
             step=step, index=index)

    design = chip.top()
    chip.set('tool', tool, 'task', task, 'output', [design + '.dot', design + '.png'],
             step=step, index=index)


################################
# format liberty files for yosys
################################
def prepare_asic_libraries(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    # Clear in case of rerun
    for libtype in ('synthesis_libraries', 'synthesis_libraries_macros'):
        chip.set('tool', tool, 'task', task, 'file', libtype, [],
                 step=step, index=index)

    # Generate synthesis_libraries and synthesis_macro_libraries for Yosys use

    # mark libs with dont_use since ABC cannot get this information via its commands
    # this also ensures the liberty files have been decompressed and corrected formatting
    # issues that generally cannot be handled by yosys or yosys-abc
    def get_synthesis_libraries(lib):
        keypath = _get_synthesis_library_key(chip, lib)
        if keypath and chip.valid(*keypath):
            return chip.find_files(*keypath, step=step, index=index)
        return []

    for libtype in ('logic', 'macro'):
        for lib in get_libraries(chip, libtype):
            lib_content = {}
            # Mark dont use
            for lib_file in get_synthesis_libraries(lib):
                # Ensure a unique name is used for library
                lib_file_name_base = os.path.basename(lib_file)
                if lib_file_name_base.lower().endswith('.gz'):
                    lib_file_name_base = lib_file_name_base[0:-3]
                if lib_file_name_base.lower().endswith('.lib'):
                    lib_file_name_base = lib_file_name_base[0:-4]

                lib_file_name = lib_file_name_base
                unique_ident = 0
                while lib_file_name in lib_content:
                    lib_file_name = f'{lib_file_name_base}_{unique_ident}'
                    unique_ident += 1

                lib_content[lib_file_name] = prepareLib.processLibertyFile(
                        lib_file,
                        logger=None if chip.get('option', 'quiet',
                                                step=step, index=index) else chip.logger)

            if not lib_content:
                continue

            var_name = 'synthesis_libraries'
            if libtype == "macro":
                var_name = 'synthesis_libraries_macros'

            for file, content in lib_content.items():
                output_file = os.path.join(
                    chip.getworkdir(step=step, index=index),
                    'inputs',
                    f'sc_{libtype}_{lib}_{file}.lib'
                )

                with open(output_file, 'w') as f:
                    f.write(content)

                chip.add('tool', tool, 'task', task, 'file', var_name, output_file,
                         step=step, index=index)


def get_synthesis_corner(chip):
    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    syn_corners = chip.get('tool', tool, 'task', task, 'var', 'synthesis_corner',
                           step=step, index=index)
    if syn_corners:
        return syn_corners

    # determine corner based on setup corner from constraints
    corner = None
    for constraint in chip.getkeys('constraint', 'timing'):
        checks = chip.get('constraint', 'timing', constraint, 'check', step=step, index=index)
        if "setup" in checks and not corner:
            corner = chip.get('constraint', 'timing', constraint, 'libcorner',
                              step=step, index=index)

    if not corner:
        # try getting it from first constraint with a valid libcorner
        for constraint in chip.getkeys('constraint', 'timing'):
            if not corner:
                corner = chip.get('constraint', 'timing', constraint, 'libcorner',
                                  step=step, index=index)

    return corner


def _get_synthesis_library_key(chip, lib):
    if chip.valid('library', lib, 'option', 'file', 'yosys_synthesis_libraries'):
        return ('library', lib, 'option', 'file', 'yosys_synthesis_libraries')

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    delaymodel = chip.get('asic', 'delaymodel', step=step, index=index)

    for corner in chip.getkeys('library', lib, 'output'):
        if chip.valid('library', lib, 'output', corner, delaymodel):
            return ('library', lib, 'output', corner, delaymodel)

    return None


##################################################
def pre_process(chip):
    ''' Tool specific function to run before step execution
    '''

    prepare_asic_libraries(chip)



from siliconcompiler import Chip
from siliconcompiler.tools.opensta import report_libraries


def main():
    chip = Chip('report_libs')
    args = chip.create_cmdline(switchlist=['-target'])

    if 'target' not in args or not args['target']:
        raise ValueError('-target is required')

    chip.node('report_lib', 'report', report_libraries)
    chip.set('option', 'flow', 'report_lib')
    chip.set('option', 'clean', True)

    chip.run()

    return 0


if __name__ == "__main__":
    main()



'''
Vivado is an FPGA programming tool suite from Xilinx used to
program Xilinx devices.

Documentation: https://www.amd.com/en/products/software/adaptive-socs-and-fpgas/vivado.html
'''

import json
import os
import re
from siliconcompiler import sc_open
from siliconcompiler.tools._common import record_metric
from siliconcompiler.targets import fpgaflow_demo


def make_docs(chip):
    chip.set('fpga', 'partname', 'ice40up5k-sg48')
    chip.use(fpgaflow_demo)


tool = 'vivado'


def setup(chip):
    vendor = 'xilinx'

    chip.set('tool', tool, 'exe', tool)
    chip.set('tool', tool, 'vendor', vendor)
    chip.set('tool', tool, 'vswitch', '-version')
    chip.set('tool', tool, 'format', 'tcl')

    # report_design_analysis -json flag requires Vivado 2021 or greater
    chip.set('tool', tool, 'version', '>=2021', clobber=False)


def setup_task(chip, task):
    setup(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    script = 'sc_run.tcl'

    refdir = f'tools/{tool}/scripts'
    option = ['-nolog', '-nojournal', '-mode', 'batch', '-source']

    chip.set('tool', tool, 'task', task, 'refdir', refdir, step=step, index=index,
             package='siliconcompiler', clobber=False)
    chip.set('tool', tool, 'task', task, 'script', script, step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'option', option, step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'regex', 'errors', r'^ERROR:',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'regex', 'warnings', r'^(CRITICAL )?WARNING:',
             step=step, index=index, clobber=False)


def parse_version(stdout):
    # Vivado v2021.2 (64-bit)
    return stdout.split()[1]


def normalize_version(version):
    return version.lstrip('v')


def _parse_qor_summary(chip, step, index):
    if not os.path.isfile('qor_summary.json'):
        return

    with sc_open('qor_summary.json') as f:
        data = json.load(f)

        # Data is organized as list of tasks that Vivado has completed, with
        # metrics associated with each. The tasks appear to be in chronological
        # order, so we pull metrics from the last one.
        task = data['Design QoR Summary'][-1]
        setup_wns = task['Wns(ns)']
        setup_tns = task['Tns(ns)']
        hold_wns = task['Whs(ns)']
        hold_tns = task['Ths(ns)']

        if setup_wns:
            record_metric(chip, step, index, 'setupwns', setup_wns,
                          'qor_summary.json',
                          source_unit='ns')
        if setup_tns:
            record_metric(chip, step, index, 'setuptns', setup_tns,
                          'qor_summary.json',
                          source_unit='ns')
        if hold_wns:
            record_metric(chip, step, index, 'holdwns', hold_wns,
                          'qor_summary.json',
                          source_unit='ns')
        if hold_tns:
            record_metric(chip, step, index, 'holdtns', hold_tns,
                          'qor_summary.json',
                          source_unit='ns')


def _parse_utilization(chip, step, index):
    if not os.path.isfile('reports/total_utilization.rpt'):
        return

    with sc_open('reports/total_utilization.rpt') as f:
        regexes = {
            'luts': (re.compile(r'(?:CLB|Slice) LUTs\*?\s+\|\s+(\d+)'), int),
            'regs': (re.compile(r'(?:CLB|Slice) Registers\s+\|\s+(\d+)'), int),
            'bram': (re.compile(r'Block RAM Tile\s+\|\s+(\d+(.\d+)?)'), float),
            # TODO: should URAM be float?
            'uram': (re.compile(r'URAM\s+\|\s+(\d+(.\d+)?)'), float)
        }
        vals = {}

        for line in f:
            for metric, (regex, datatype) in regexes.items():
                if metric in vals:
                    continue
                match = regex.search(line)
                if match:
                    vals[metric] = datatype(match.group(1))
                    continue

        if 'luts' in vals:
            record_metric(chip, step, index, 'luts', vals['luts'],
                          'reports/total_utilization.rpt')
        if 'regs' in vals:
            record_metric(chip, step, index, 'registers', vals['regs'],
                          'reports/total_utilization.rpt')

        total_bram = 0
        if 'bram' in vals:
            total_bram += vals['bram']
        if 'uram' in vals:
            total_bram += vals['uram']
        if 'bram' in vals or 'uram' in vals:
            record_metric(chip, step, index, 'brams', total_bram,
                          'reports/total_utilization.rpt')


def post_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    _parse_qor_summary(chip, step, index)
    _parse_utilization(chip, step, index)



from .. import _common


def get_mainlib(chip):
    return get_libraries(chip, 'logic')[0]


def get_libraries(chip, type):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    if type not in ('logic', 'macro'):
        raise ValueError(f'Cannot collect {type} libraries')

    libs = []
    for lib in chip.get('asic', f'{type}lib', step=step, index=index):
        if lib in libs:
            continue
        libs.append(lib)

    for lib in _common.get_libraries(chip, include_asic=False):
        if not chip.valid('library', lib, 'asic', f'{type}lib'):
            continue
        for sublib in chip.get('library', lib, 'asic', f'{type}lib', step=step, index=index):
            if sublib in libs:
                continue
            libs.append(sublib)

    return libs


def get_timing_modes(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    modes = set()
    for scenario in chip.getkeys('constraint', 'timing'):
        modes.add(chip.get('constraint', 'timing', scenario, 'mode',
                  step=step, index=index))
    return sorted(modes)


def set_tool_task_var(chip,
                      param_key,
                      default_value=None,
                      schelp=None,
                      option_key=None,
                      pdk_key=None,
                      lib_key=None,
                      require=None,
                      skip=None):
    '''
    Set parameter from PDK -> main library -> option -> default_value
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = _common.get_tool_task(chip, step, index)
    pdkname = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')

    if not skip:
        skip = []
    if not isinstance(skip, (list, tuple)):
        skip = [skip]

    if not require:
        require = []
    if not isinstance(require, (list, tuple)):
        require = [require]

    check_keys = []

    # Add PDK key
    if 'pdk' not in skip:
        if not pdk_key:
            pdk_key = param_key
        check_keys.append(['pdk', pdkname, 'var', tool, stackup, pdk_key])
        if 'pdk' in require:
            chip.add('tool', tool, 'task', task, 'require',
                     ','.join(check_keys[-1]),
                     step=step, index=index)

    # Add library key
    if 'lib' not in skip:
        mainlib = get_mainlib(chip)
        if not lib_key:
            lib_key = f'{tool}_{param_key}'
        check_keys.append(['library', mainlib, 'option', 'var', lib_key])
        if 'lib' in require:
            chip.add('tool', tool, 'task', task, 'require',
                     ','.join(check_keys[-1]),
                     step=step, index=index)

    # Add option key
    if 'option' not in skip:
        if not option_key:
            option_key = f'{tool}_{param_key}'
        check_keys.append(['option', 'var', option_key])
        if 'option' in require:
            chip.add('tool', tool, 'task', task, 'require',
                     ','.join(check_keys[-1]),
                     step=step, index=index)

    require_key, value = _common.pick_key(chip, reversed(check_keys), step=step, index=index)
    if not value:
        value = default_value

    if value:
        chip.set('tool', tool, 'task', task, 'var', param_key, value,
                 step=step, index=index, clobber=False)

        if require_key:
            chip.add('tool', tool, 'task', task, 'require',
                     ','.join(require_key),
                     step=step, index=index)

    if value or 'key' in require:
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['tool', tool, 'task', task, 'var', param_key]),
                 step=step, index=index)

    if schelp:
        chip.set('tool', tool, 'task', task, 'var', param_key,
                 schelp, field='help')

    return value


def get_tool_task_var(chip,
                      param_key,
                      option_key=None,
                      pdk_key=None,
                      lib_key=None,
                      skip=None):
    '''
    Get parameter from PDK -> main library -> option -> default_value
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, _ = _common.get_tool_task(chip, step, index)
    pdkname = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')

    if not skip:
        skip = []
    if not isinstance(skip, (list, tuple)):
        skip = [skip]

    check_keys = []
    # Add PDK key
    if 'pdk' not in skip:
        if not pdk_key:
            pdk_key = param_key
        check_keys.append(['pdk', pdkname, 'var', tool, stackup, pdk_key])

    # Add library key
    if 'lib' not in skip:
        mainlib = get_mainlib(chip)
        if not lib_key:
            lib_key = f'{tool}_{param_key}'
        check_keys.append(['library', mainlib, 'option', 'var', lib_key])

    # Add option key
    if 'option' not in skip:
        if not option_key:
            option_key = f'{tool}_{param_key}'
        check_keys.append(['option', 'var', option_key])

    _, value = _common.pick_key(chip, reversed(check_keys), step=step, index=index)

    return value




from siliconcompiler.tools.openroad.openroad import setup as setup_tool
from siliconcompiler.tools.openroad.openroad import build_pex_corners
from siliconcompiler.tools.openroad.openroad import post_process as or_post_process
from siliconcompiler.tools.openroad.openroad import pre_process as or_pre_process
from siliconcompiler.tools.openroad.openroad import _set_reports, set_pnr_inputs, set_pnr_outputs


def setup(chip):
    '''
    Perform global and detail placements along with design violation repairs
    '''

    # Generic tool setup.
    setup_tool(chip)

    set_pnr_inputs(chip)
    set_pnr_outputs(chip)

    _set_reports(chip, [
        'setup',
        'unconstrained',
        'power',
        'drv_violations',
        'fmax',

        # Images
        'placement_density',
        'routing_congestion',
        'power_density',
        'optimization_placement'
    ])


def pre_process(chip):
    or_pre_process(chip)
    build_pex_corners(chip)


def post_process(chip):
    or_post_process(chip)



# Version number following semver standard.
version = '0.28.7'

# Default server address for remote runs, if unspecified.
default_server = 'https://server.siliconcompiler.com'

# This is the list of significant contributors to SiliconCompiler in
# chronological order.
#
# This does not necessarily list everyone who has contributed code,
# especially since many employees of one corporation may be contributing.
# To see the full list of contributors, see the git revision history
authors = [
    'Andreas Olofsson',
    'William Ransohoff',
    'Noah Moroze',
    'Zachary Yedidia',
    'Massimiliano Giacometti',
    'Kimia Talaei',
    'Peter Gadfort',
    'Aulihan Teng',
    'Peter Grossmann',
    'Gabriel Aguirre',
    'Martin Troiber'
]

# CLI entry banner autogenerated using pyfiglet.
# >> pyfiglet.figlet_format("Silicon Compiler")
banner = r'''
 ____  _ _ _                    ____                      _ _
/ ___|(_) (_) ___ ___  _ __    / ___|___  _ __ ___  _ __ (_) | ___ _ __
\___ \| | | |/ __/ _ \| '_ \  | |   / _ \| '_ ` _ \| '_ \| | |/ _ \ '__|
 ___) | | | | (_| (_) | | | | | |__| (_) | | | | | | |_) | | |  __/ |
|____/|_|_|_|\___\___/|_| |_|  \____\___/|_| |_| |_| .__/|_|_|\___|_|
                                                   |_|
'''



import siliconcompiler

from siliconcompiler.tools.surelog import parse as surelog_parse
from siliconcompiler.tools.yosys import syn_asic
from siliconcompiler.tools.klayout import export
from siliconcompiler.flows._common import _make_docs


def setup():
    '''A flow for stitching together hardened blocks without doing any automated
    place-and-route.

    This flow generates a GDS and a netlist for passing to a
    verification/signoff flow.
    '''
    flow = siliconcompiler.Flow('asictopflow')

    flow.node(flow.design, 'import', surelog_parse)
    flow.node(flow.design, 'syn', syn_asic)
    flow.node(flow.design, 'export', export)

    flow.edge(flow.design, 'import', 'export')
    flow.edge(flow.design, 'import', 'syn')

    # Set default goal
    for step in flow.getkeys('flowgraph', flow.design):
        flow.set('flowgraph', flow.design, step, '0', 'goal', 'errors', 0)

    return flow


##################################################
if __name__ == "__main__":
    chip = siliconcompiler.Chip('design')
    _make_docs(chip)
    flow = setup()
    chip.use(flow)
    chip.write_flowgraph(f"{flow.top()}.png", flow=flow.top())



from siliconcompiler.tools.openroad import openroad
from siliconcompiler.tools._common import get_tool_task


def setup_tool(chip):
    ''' Helper method for configs specific to extraction tasks.
    '''

    openroad.setup(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'script', 'sc_rcx.tcl', step=step, index=index)
    chip.set('tool', tool, 'task', task, 'threads', 1, step=step, index=index)


def setup_task(chip):
    # Generic tool setup.
    setup_tool(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    pdk = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')

    chip.set('tool', tool, 'task', task, 'var', 'libtype',
             list(chip.getkeys('pdk', pdk, 'aprtech', tool, stackup))[0],
             clobber=False, step=step, index=index)
    chip.set('tool', tool, 'task', task, 'var', 'libtype',
             'Library type used to select the lef file',
             field='help')
    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['tool', tool, 'task', task, 'var', 'libtype']),
             step=step, index=index)
    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['pdk',
                       pdk,
                       'aprtech',
                       tool,
                       stackup,
                       chip.get('tool', tool, 'task', task, 'var', 'libtype',
                                step=step, index=index)[0],
                       'lef']),
             step=step, index=index)


def setup(chip):
    ''' Helper method for configs specific to extraction tasks.
    '''

    # Generic tool setup.
    setup_task(chip)

    design = chip.top()

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.add('tool', tool, 'task', task, 'output', f'{design}.def', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', f'{design}.vg', step=step, index=index)

    pdk = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')

    if chip.valid('pdk', pdk, 'var', 'openroad', 'rcx_bench_max_layer', stackup):
        chip.set('tool', tool, 'task', task, 'var', 'max_layer',
                 chip.get('pdk', pdk, 'var', 'openroad', 'rcx_bench_max_layer', stackup)[0],
                 clobber=False, step=step, index=index)
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['pdk', pdk, 'var', 'openroad', 'rcx_bench_max_layer', stackup]),
                 step=step, index=index)
    else:
        chip.set('tool', tool, 'task', task, 'var', 'max_layer',
                 chip.get('pdk', pdk, 'maxlayer', stackup),
                 clobber=False, step=step, index=index)
    chip.set('tool', tool, 'task', task, 'var', 'max_layer',
             'Maximum layer to generate extraction bench for',
             field='help')
    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['tool', tool, 'task', task, 'var', 'max_layer']),
             step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'bench_length', '100',
             clobber=False, step=step, index=index)
    chip.set('tool', tool, 'task', task, 'var', 'bench_length',
             'Length of bench wires',
             field='help')
    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['tool', tool, 'task', task, 'var', 'bench_length']),
             step=step, index=index)



from siliconcompiler.tools._common import input_provides, input_file_node_name, get_tool_task

from siliconcompiler.tools.klayout import klayout
from siliconcompiler.tools.klayout.klayout import setup as setup_tool


def make_docs(chip):
    klayout.make_docs(chip)


def setup(chip):
    '''
    Convert a DRC db from .lyrdb or .ascii to an openroad json marker file
    '''

    # Generic tool setup.
    setup_tool(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    design = chip.top()

    clobber = False

    chip.set('tool', tool, 'task', task, 'threads', 1,
             step=step, index=index, clobber=clobber)

    script = 'klayout_convert_drc_db.py'
    option = ['-z', '-nc', '-rx', '-r']
    chip.set('tool', tool, 'task', task, 'script', script,
             step=step, index=index, clobber=clobber)
    chip.set('tool', tool, 'task', task, 'option', option,
             step=step, index=index, clobber=clobber)

    input_nodes = set()
    for nodes in input_provides(chip, step, index).values():
        input_nodes.update(nodes)

    chip.set('tool', tool, 'task', task, 'input', [], step=step, index=index)
    chip.set('tool', tool, 'task', task, 'output', [], step=step, index=index)
    for file, nodes in input_provides(chip, step, index).items():
        if file not in (f'{design}.lyrdb', f'{design}.ascii'):
            continue

        if len(input_nodes) == 1:
            chip.add('tool', tool, 'task', task, 'input',
                     file,
                     step=step, index=index)
        else:
            for in_step, in_index in nodes:
                chip.add('tool', tool, 'task', task, 'input',
                         input_file_node_name(file, in_step, in_index),
                         step=step, index=index)
    if not chip.get('tool', tool, 'task', task, 'input', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'input', f'{design}.lyrdb',
                 step=step, index=index)

    chip.set('tool', tool, 'task', task, 'output', f'{design}.json',
             step=step, index=index)



# Copyright 2022 Silicon Compiler Authors. All Rights Reserved.

# NOTE: this file cannot rely on any third-party dependencies, including other
# SC dependencies outside of its directory, since it may be used by tool drivers
# that have isolated Python environments.

import copy
import logging
import os
import re
import pathlib
import argparse
import sys
import shlex

try:
    import orjson as json
    _has_orjson = True
except ModuleNotFoundError:
    import json
    _has_orjson = False

try:
    import gzip
    _has_gzip = True
except ModuleNotFoundError:
    _has_gzip = False

try:
    import csv
    _has_csv = True
except ModuleNotFoundError:
    _has_csv = False

try:
    import yaml
    _has_yaml = True
except ImportError:
    _has_yaml = False

from .schema_cfg import schema_cfg
from .utils import escape_val_tcl, PACKAGE_ROOT, translate_loglevel


class Schema:
    """Object for storing and accessing configuration values corresponding to
    the SiliconCompiler schema.

    Most user-facing interaction with the schema should occur through an
    instance of :class:`~siliconcompiler.core.Chip`, but this class is available
    for schema manipulation tasks that don't require the additional context of a
    Chip object.

    The two arguments to this class are mutually exclusive. If neither are
    provided, the object is initialized to default values for all parameters.

    Args:
        cfg (dict): Initial configuration dictionary. This may be a subtree of
            the schema.
        manifest (str): Initial manifest.
        logger (logging.Logger): instance of the parent logger if available
    """

    _RECORD_ACCESS_IDENTIFIER = "SC_CFG_ACCESS_KEY"

    # Special key in node dict that represents a value corresponds to a
    # global default for all steps/indices.
    GLOBAL_KEY = 'global'
    PERNODE_FIELDS = ('value', 'filehash', 'date', 'author', 'signature', 'package')

    def __init__(self, cfg=None, manifest=None, logger=None):
        if cfg is not None and manifest is not None:
            raise ValueError('You may not specify both cfg and manifest')

        # Use during testing to record calls to Schema.get
        self._init_record_access()

        self._init_logger(logger)

        self._stop_journal()

        if manifest is not None:
            # Normalize value to string in case we receive a pathlib.Path
            cfg, self.__journal = Schema.__read_manifest_file(str(manifest))
        else:
            cfg = copy.deepcopy(cfg)

        if cfg is not None:
            try:
                if Schema.__dict_requires_normalization(cfg):
                    cfg = Schema.__dict_to_schema(cfg)
                self.cfg = cfg
            except (TypeError, ValueError) as e:
                raise ValueError('Attempting to read manifest with '
                                 f'incompatible schema version: {e}') \
                    from e
        else:
            self.cfg = self._init_schema_cfg()

    ###########################################################################
    def _init_schema_cfg(self):
        return schema_cfg()

    ###########################################################################
    @staticmethod
    def __dict_to_schema_set(cfg, *key):
        if Schema._is_leaf(cfg):
            for field, value in cfg.items():
                if field == 'node':
                    for step, substep in value.items():
                        if step == 'default':
                            continue
                        for index, values in substep.items():
                            if step == Schema.GLOBAL_KEY:
                                sstep = None
                            else:
                                sstep = step
                            if index == Schema.GLOBAL_KEY:
                                sindex = None
                            else:
                                sindex = index
                            for nodefield, nodevalue in values.items():
                                Schema.__set(*key, nodevalue,
                                             cfg=cfg,
                                             field=nodefield,
                                             step=sstep, index=sindex)
                else:
                    Schema.__set(*key, value, cfg=cfg, field=field)
        else:
            for nextkey, subcfg in cfg.items():
                Schema.__dict_to_schema_set(subcfg, *key, nextkey)

    ###########################################################################
    @staticmethod
    def __dict_to_schema(cfg):
        for category, subcfg in cfg.items():
            if category in ('history', 'library'):
                # History and library are subschemas
                for _, value in subcfg.items():
                    Schema.__dict_to_schema(value)
            else:
                Schema.__dict_to_schema_set(subcfg, category)
        return cfg

    ###########################################################################
    @staticmethod
    def __dict_requires_normalization(cfg):
        '''
        Recurse over scheme configuration to check for tuples
        Returns: False if dict is correct, True is dict requires normalization,
            None if tuples were not found
        '''
        if Schema._is_leaf(cfg):
            if '(' in cfg['type']:
                for step, substep in cfg['node'].items():
                    for index, values in substep.items():
                        values = values['value']
                        if not values:
                            continue
                        if isinstance(values, list):
                            for v in values:
                                if isinstance(v, tuple):
                                    return False
                        if isinstance(values, tuple):
                            return False
                        return True
            else:
                return None
        else:
            for subcfg in cfg.values():
                ret = Schema.__dict_requires_normalization(subcfg)
                if ret is None:
                    continue
                else:
                    return ret

    ###########################################################################
    def _merge_with_init_schema(self):
        new_schema = Schema()

        for keylist in self.allkeys():
            if keylist[0] in ('history', 'library'):
                continue

            if 'default' in keylist:
                continue

            # only read in valid keypaths without 'default'
            key_valid = new_schema.valid(*keylist, default_valid=True)
            if not key_valid:
                self.logger.warning(f'Keypath {keylist} is not valid')
            if not key_valid:
                continue

            for val, step, index in self._getvals(*keylist, return_defvalue=False):
                new_schema.set(*keylist, val, step=step, index=index)

                # update other pernode fields
                # TODO: only update these if clobber is successful
                step_key = Schema.GLOBAL_KEY if not step else step
                idx_key = Schema.GLOBAL_KEY if not index else index
                for field in self.getdict(*keylist)['node'][step_key][idx_key].keys():
                    if field == 'value':
                        continue
                    new_schema.set(*keylist,
                                   self.get(*keylist, step=step, index=index, field=field),
                                   step=step, index=index, field=field)

        if 'library' in self.cfg:
            # Handle libraries separately
            for library in self.cfg['library'].keys():
                lib_schema = Schema(cfg=self.getdict('library', library))
                lib_schema._merge_with_init_schema()
                new_schema.cfg['library'][library] = lib_schema.cfg

        if 'history' in self.cfg:
            # Copy over history
            new_schema.cfg['history'] = self.cfg['history']

        self.cfg = new_schema.cfg

    ###########################################################################
    @staticmethod
    def __read_manifest_file(filepath):
        if not os.path.isfile(filepath):
            raise ValueError(f'Manifest file not found {filepath}')

        if os.path.splitext(filepath)[1].lower() == '.gz':
            if not _has_gzip:
                raise RuntimeError("gzip is not available")
            fin = gzip.open(filepath, 'r')
        else:
            fin = open(filepath, 'r')

        try:
            if re.search(r'(\.json|\.sup)(\.gz)*$', filepath, flags=re.IGNORECASE):
                localcfg = json.loads(fin.read())
            elif re.search(r'(\.yaml|\.yml)(\.gz)*$', filepath, flags=re.IGNORECASE):
                if not _has_yaml:
                    raise ImportError('yaml package required to read YAML manifest')
                localcfg = yaml.load(fin, Loader=yaml.SafeLoader)
            else:
                raise ValueError(f'File format not recognized {filepath}')
        finally:
            fin.close()

        journal = None
        try:
            if '__journal__' in localcfg:
                journal = localcfg['__journal__']
                del localcfg['__journal__']
        except (TypeError, ValueError) as e:
            raise ValueError(f'Attempting to read manifest with incompatible schema version: {e}') \
                from e

        return localcfg, journal

    def get(self, *keypath, field='value', job=None, step=None, index=None):
        """
        Returns a schema parameter field.

        See :meth:`~siliconcompiler.core.Chip.get` for detailed documentation.
        """

        if self.__record_access["recording"]:
            self.__record_access["record"].add(tuple(keypath))

        # Prevent accidental modifications of the schema content by not passing a reference
        return copy.copy(self.__get(*keypath, field=field, job=job, step=step, index=index))

    ###########################################################################
    def __get(self, *keypath, field='value', job=None, step=None, index=None):
        cfg = self.__search(*keypath, job=job)

        if not Schema._is_leaf(cfg):
            raise ValueError(f'Invalid keypath {keypath}: get() '
                             'must be called on a complete keypath')

        err = Schema.__validate_step_index(cfg['pernode'], field, step, index)
        if err:
            raise ValueError(f'Invalid args to get() of keypath {keypath}: {err}')

        if isinstance(index, int):
            index = str(index)

        if field in self.PERNODE_FIELDS:
            try:
                return cfg['node'][step][index][field]
            except KeyError:
                if cfg['pernode'] == 'required':
                    return cfg['node']['default']['default'][field]

            try:
                return cfg['node'][step][self.GLOBAL_KEY][field]
            except KeyError:
                pass

            try:
                return cfg['node'][self.GLOBAL_KEY][self.GLOBAL_KEY][field]
            except KeyError:
                return cfg['node']['default']['default'][field]
        elif field in cfg:
            return cfg[field]
        else:
            raise ValueError(f'Invalid field {field}')

    ###########################################################################
    def set(self, *args, field='value', clobber=True, step=None, index=None):
        '''
        Sets a schema parameter field.

        See :meth:`~siliconcompiler.core.Chip.set` for detailed documentation.
        '''

        keypath = args[:-1]
        cfg = self.__search(*keypath, insert_defaults=True)

        return self.__set(*args, logger=self.logger, cfg=cfg, field=field, clobber=clobber,
                          step=step, index=index, journal_callback=self.__record_journal)

    ###########################################################################
    @staticmethod
    def __set(*args, logger=None, cfg=None, field='value', clobber=True,
              step=None, index=None,
              journal_callback=None):
        '''
        Sets a schema parameter field.

        See :meth:`~siliconcompiler.core.Chip.set` for detailed documentation.
        '''
        keypath = args[:-1]
        value = args[-1]

        if not Schema._is_leaf(cfg):
            raise ValueError(f'Invalid keypath {keypath}: set() '
                             'must be called on a complete keypath')

        err = Schema.__validate_step_index(cfg['pernode'], field, step, index)
        if err:
            raise ValueError(f'Invalid args to set() of keypath {keypath}: {err}')

        if isinstance(index, int):
            index = str(index)

        if cfg['lock'] and field != 'lock':
            if logger:
                logger.debug(f'Failed to set value for {keypath}: parameter is locked')
            return False

        if Schema.__is_set(cfg, step=step, index=index) and not clobber:
            if logger:
                logger.debug(f'Failed to set value for {keypath}: clobber is False '
                             'and parameter is set')
            return False

        allowed_values = None
        if 'enum' in cfg:
            allowed_values = cfg['enum']

        value = Schema.__check_and_normalize(value, cfg['type'], field, keypath, allowed_values)

        if journal_callback:
            journal_callback("set", keypath,
                             value=value,
                             field=field,
                             step=step, index=index)

        if field in Schema.PERNODE_FIELDS:
            step = step if step is not None else Schema.GLOBAL_KEY
            index = index if index is not None else Schema.GLOBAL_KEY

            if step not in cfg['node']:
                cfg['node'][step] = {}
            if index not in cfg['node'][step]:
                cfg['node'][step][index] = copy.deepcopy(cfg['node']['default']['default'])
            cfg['node'][step][index][field] = value
        else:
            cfg[field] = value

        return True

    ###########################################################################
    def add(self, *args, field='value', step=None, index=None):
        '''
        Adds item(s) to a schema parameter list.

        See :meth:`~siliconcompiler.core.Chip.add` for detailed documentation.
        '''
        keypath = args[:-1]

        cfg = self.__search(*keypath, insert_defaults=True)

        return self._add(*args, cfg=cfg, field=field, step=step, index=index)

    ###########################################################################
    def _add(self, *args, cfg=None, field='value', step=None, index=None, package=None):
        '''
        Adds item(s) to a schema parameter list.

        See :meth:`~siliconcompiler.core.Chip.add` for detailed documentation.
        '''
        keypath = args[:-1]
        value = args[-1]

        if not Schema._is_leaf(cfg):
            raise ValueError(f'Invalid keypath {keypath}: add() '
                             'must be called on a complete keypath')

        err = Schema.__validate_step_index(cfg['pernode'], field, step, index)
        if err:
            raise ValueError(f'Invalid args to add() of keypath {keypath}: {err}')

        if isinstance(index, int):
            index = str(index)

        if not Schema.__is_list(field, cfg['type']):
            if field == 'value':
                raise ValueError(f'Invalid keypath {keypath}: add() must be called on a list')
            else:
                raise ValueError(f'Invalid field {field}: add() must be called on a list')

        if cfg['lock']:
            self.logger.debug(f'Failed to set value for {keypath}: parameter is locked')
            return False

        allowed_values = None
        if 'enum' in cfg:
            allowed_values = cfg['enum']

        value = Schema.__check_and_normalize(value, cfg['type'], field, keypath, allowed_values)
        self.__record_journal("add", keypath, value=value, field=field, step=step, index=index)
        if field in self.PERNODE_FIELDS:
            modified_step = step if step is not None else self.GLOBAL_KEY
            modified_index = index if index is not None else self.GLOBAL_KEY

            if modified_step not in cfg['node']:
                cfg['node'][modified_step] = {}
            if modified_index not in cfg['node'][modified_step]:
                cfg['node'][modified_step][modified_index] = copy.deepcopy(
                    cfg['node']['default']['default'])
            cfg['node'][modified_step][modified_index][field].extend(value)
        else:
            cfg[field].extend(value)

        return True

    ###########################################################################
    def change_type(self, *key, type=None):
        '''
        Change the type of a key

        Args:
            key (list): Key to change.
            type (str): New data type for this key

        Examples:
            >>> chip.set('option', 'var', 'run_test', 'true')
            >>> chip.schema.change_type('option', 'var', 'run_test', 'bool')
            Changes the type of ['option', 'var', 'run_test'] to a boolean.
        '''

        if not type:
            raise ValueError('Type cannot be empty')

        if 'file' in type or 'dir' in type:
            raise ValueError(f'Cannot convert to {type}')

        cfg = self.__search(*key, insert_defaults=True)
        if not Schema._is_leaf(cfg):
            raise ValueError(f'Invalid keypath {key}: change_type() '
                             'must be called on a complete keypath')

        old_type = self.get(*key, field='type')
        if 'file' in old_type or 'dir' in old_type:
            raise ValueError(f'Cannot convert from {old_type}')

        old_type_is_list = '[' in old_type
        new_type_is_list = '[' in type

        if 'file' in old_type or 'dir' in old_type:
            raise ValueError(f'Cannot convert from {type}')

        new_values = []
        for values, step, index in [*self._getvals(*key),
                                    (self.get_default(*key), 'default', 'default')]:
            if old_type_is_list and not new_type_is_list:
                # Old type is list, but new type in not a list
                # Can only convert if list has 1 or 0 elements
                if len(values) > 1:
                    raise ValueError(f'Too many values in {",".join(key)} to convert a '
                                     'list of a scalar.')
                if len(values) == 1:
                    values = values[0]
                else:
                    values = None

            if new_type_is_list and values is None:
                values = []

            new_values.append((step, index, values))

        self.set(*key, type, field='type')
        for step, index, values in new_values:
            if step == 'default' and index == 'default':
                self.set_default(*key, values)
            else:
                self.set(*key, values, step=step, index=index)

    ###########################################################################
    def remove(self, *keypath):
        '''
        Remove a keypath

        See :meth:`~siliconcompiler.core.Chip.remove` for detailed documentation.
        '''
        search_path = keypath[0:-1]
        removal_key = keypath[-1]

        if removal_key == 'default':
            self.logger.error(f'Cannot remove default keypath: {keypath}')
            return

        cfg = self.__search(*search_path)
        if 'default' not in cfg:
            self.logger.error(f'Cannot remove a non-default keypath: {keypath}')
            return

        if removal_key not in cfg:
            self.logger.error(f'Key does not exist: {keypath}')
            return

        for key in self.allkeys(*keypath):
            fullpath = [*keypath, *key]
            if self.get(*fullpath, field='lock'):
                self.logger.error(f'Key is locked: {fullpath}')
                return

        del cfg[removal_key]
        self.__record_journal("remove", keypath)

    ###########################################################################
    def unset(self, *keypath, step=None, index=None):
        '''
        Unsets a schema parameter field.

        See :meth:`~siliconcompiler.core.Chip.unset` for detailed documentation.
        '''
        cfg = self.__search(*keypath)

        if not Schema._is_leaf(cfg):
            raise ValueError(f'Invalid keypath {keypath}: unset() '
                             'must be called on a complete keypath')

        err = Schema.__validate_step_index(cfg['pernode'], 'value', step, index)
        if err:
            raise ValueError(f'Invalid args to unset() of keypath {keypath}: {err}')

        if isinstance(index, int):
            index = str(index)

        if cfg['lock']:
            self.logger.debug(f'Failed to set value for {keypath}: parameter is locked')
            return False

        if step is None:
            step = Schema.GLOBAL_KEY
        if index is None:
            index = Schema.GLOBAL_KEY

        try:
            del cfg['node'][step][index]
            self.__record_journal("unset", keypath, step=step, index=index)
        except KeyError:
            # If this key doesn't exist, silently continue - it was never set
            pass

        return True

    def _getvals(self, *keypath, return_defvalue=True):
        """
        Returns all values (global and pernode) associated with a particular parameter.

        Returns a list of tuples of the form (value, step, index). The list is
        in no particular order. For the global value, step and index are None.
        If return_defvalue is True, the default parameter value is added to the
        list in place of a global value if a global value is not set.
        """
        cfg = self.__search(*keypath)

        if not Schema._is_leaf(cfg):
            raise ValueError(f'Invalid keypath {keypath}: _getvals() '
                             'must be called on a complete keypath')

        vals = []
        has_global = False
        for step in cfg['node']:
            if step == 'default':
                continue

            for index in cfg['node'][step]:
                step_arg = None if step == self.GLOBAL_KEY else step
                index_arg = None if index == self.GLOBAL_KEY else index
                if 'value' in cfg['node'][step][index]:
                    if step_arg is None and index_arg is None:
                        has_global = True
                    vals.append((cfg['node'][step][index]['value'], step_arg, index_arg))

        if (cfg['pernode'] != 'required') and not has_global and return_defvalue:
            vals.append((cfg['node']['default']['default']['value'], None, None))

        return vals

    ###########################################################################
    def getkeys(self, *keypath, job=None):
        """
        Returns a list of schema dictionary keys.

        See :meth:`~siliconcompiler.core.Chip.getkeys` for detailed documentation.
        """
        cfg = self.__search(*keypath, job=job, use_default=False)
        keys = list(cfg.keys())

        if 'default' in keys:
            keys.remove('default')

        return keys

    ###########################################################################
    def getdict(self, *keypath):
        """
        Returns a schema dictionary.

        See :meth:`~siliconcompiler.core.Chip.getdict` for detailed
        documentation.
        """
        cfg = self.__search(*keypath)
        return copy.deepcopy(cfg)

    ###########################################################################
    def valid(self, *args, default_valid=False, job=None, check_complete=False):
        """
        Checks validity of a keypath.

        See :meth:`~siliconcompiler.core.Chip.valid` for detailed
        documentation.
        """
        keylist = list(args)
        if default_valid:
            default = 'default'
        else:
            default = None

        if job is not None:
            cfg = self.cfg['history'][job]
        else:
            cfg = self.cfg

        for key in keylist:
            if key in cfg:
                cfg = cfg[key]
            elif default_valid and default in cfg:
                cfg = cfg[default]
            else:
                return False
        if check_complete:
            return Schema._is_leaf(cfg)
        return True

    ##########################################################################
    def has_field(self, *args):
        keypath = args[:-1]
        field = args[-1]

        cfg = self.__search(*keypath)
        return field in cfg

    ##########################################################################
    def record_history(self):
        '''
        Copies all non-empty parameters from current job into the history
        dictionary.
        '''

        # initialize new dict
        jobname = self.get('option', 'jobname')
        self.cfg['history'][jobname] = {}

        # copy in all empty values of scope job
        allkeys = self.allkeys()
        for key in allkeys:
            # ignore history in case of cumulative history
            if key[0] != 'history':
                scope = self.get(*key, field='scope')
                if not self.is_empty(*key) and (scope == 'job'):
                    self.__copyparam(self.cfg,
                                     self.cfg['history'][jobname],
                                     key)

    @staticmethod
    def __check_and_normalize(value, sc_type, field, keypath, allowed_values):
        '''
        This method validates that user-provided values match the expected type,
        and returns a normalized version of the value.

        The expected type is based on the schema parameter type string for
        value-related fields, and is based on the field itself for other fields.
        This function raises a TypeError if an illegal value is provided.

        The normalization process provides some leeway in how users supply
        values, while ensuring that values are stored consistently in the schema.

        The normalization rules are as follows:
        - If a scalar is provided for a list type, it is promoted to a list of
        one element.
        - If a list is provided for a tuple type, it is cast to a tuple (since
        the JSON module serializes tuples as arrays, which are deserialized into
        lists).
        - Elements inside lists and tuples are normalized recursively.
        - All non-list values have a string representation that gets cast to a
        native Python type (since we receive strings from the CLI):
          - bool: accepts "true" or "false"
          - ints and floats: cast as if by int() or float()
          - tuples: accepts comma-separated values surrounded by parens
        '''

        if value is None and not Schema.__is_list(field, sc_type):
            # None is legal for all scalars, but not within collection types
            # TODO: could consider normalizing "None" for lists to empty list?
            return value

        if field == 'value':
            # Push down error_msg from the top since arguments get modified in recursive call
            error_msg = f'Invalid value {value} for keypath {keypath}: expected type {sc_type}'
            return Schema._normalize_value(value, sc_type, error_msg, allowed_values)
        else:
            return Schema.__normalize_field(value, sc_type, field, keypath)

    @staticmethod
    def _normalize_value(value, sc_type, error_msg, allowed_values):
        if sc_type.startswith('['):
            base_type = sc_type[1:-1]

            # Need to try 2 different recursion strategies - if value is a list already, then we can
            # recurse on it directly. However, if that doesn't work, then it might be a
            # list-of-lists/tuples that needs to be wrapped in an outer list, so we try that.
            if isinstance(value, (list, set, tuple)):
                try:
                    return [Schema._normalize_value(v, base_type, error_msg, allowed_values)
                            for v in value]
                except TypeError:
                    pass

            value = [value]
            return [Schema._normalize_value(v, base_type, error_msg, allowed_values) for v in value]

        if sc_type.startswith('('):
            # TODO: make parsing more robust to support tuples-of-tuples
            if isinstance(value, str):
                value = value[1:-1].split(',')
            elif not (isinstance(value, tuple) or isinstance(value, list)):
                raise TypeError(error_msg)

            base_types = sc_type[1:-1].split(',')
            if len(value) != len(base_types):
                raise TypeError(error_msg)
            return tuple(Schema._normalize_value(v, base_type, error_msg, allowed_values)
                         for v, base_type in zip(value, base_types))

        if sc_type == 'bool':
            if value == 'true':
                return True
            if value == 'false':
                return False
            if isinstance(value, bool):
                return value
            if isinstance(value, (int, float)):
                return value != 0
            raise TypeError(error_msg)

        try:
            if sc_type == 'int':
                return int(value)

            if sc_type == 'float':
                return float(value)
        except TypeError:
            raise TypeError(error_msg) from None

        if sc_type == 'str':
            if isinstance(value, str):
                return value
            elif isinstance(value, bool):
                return str(value).lower()
            elif isinstance(value, (list, tuple)):
                raise TypeError(error_msg)
            else:
                return str(value)

        if sc_type in ('file', 'dir'):
            if isinstance(value, (str, pathlib.Path)):
                return str(value)
            else:
                raise TypeError(error_msg)

        if sc_type == 'enum':
            if isinstance(value, str):
                if value in allowed_values:
                    return value
                valid = ", ".join(allowed_values)
                raise ValueError(error_msg + f", and value of {valid}")
            else:
                raise TypeError(error_msg)

        raise ValueError(f'Invalid type specifier: {sc_type}')

    @staticmethod
    def __normalize_field(value, sc_type, field, keypath):
        def error_msg(t):
            return f'Invalid value {value} for field {field} of keypath {keypath}: expected {t}'

        if field in ('author', 'date') and ('file' not in sc_type):
            raise TypeError(f'Invalid field {field} for keypath {keypath}: '
                            'this field only exists for file parameters')

        if field in ('copy', 'filehash', 'package', 'hashalgo') and \
           ('file' not in sc_type and 'dir' not in sc_type):
            raise TypeError(f'Invalid field {field} for keypath {keypath}: '
                            'this field only exists for file and dir parameters')

        is_list = Schema.__is_list(field, sc_type)
        if field == 'package' and is_list:
            if not isinstance(value, list):
                value = [value]
            if not all((v is None or isinstance(v, (str, pathlib.Path))) for v in value):
                raise TypeError(error_msg('None, str or pathlib.Path'))
            return value

        if is_list:
            if not value:
                # Replace none with an empty list
                value = []

            if not isinstance(value, list):
                value = [value]

            if not all(isinstance(v, str) for v in value):
                raise TypeError(error_msg('str'))
            return value

        if field == 'scope':
            # Restricted allowed values
            if not (isinstance(value, str) and value in ('global', 'job', 'scratch')):
                raise TypeError(error_msg('one of "global", "job", or "scratch"'))
            return value

        if field == 'pernode':
            # Restricted allowed values
            if not (isinstance(value, str) and value in ('never', 'optional', 'required')):
                raise TypeError(f'Invalid value {value} for field {field}: '
                                'expected one of "never", "optional", or "required"')
            return value

        if field in (
            'type', 'switch', 'shorthelp', 'help', 'unit', 'hashalgo', 'notes',
            'signature'
        ):
            if not isinstance(value, str):
                raise TypeError(error_msg('str'))
            return value

        if field in ('lock', 'copy', 'require'):
            if value == 'true':
                return True
            if value == 'false':
                return False
            if isinstance(value, bool):
                return value
            else:
                raise TypeError(error_msg('bool'))

        if field in ('node',):
            if isinstance(value, dict):
                return value
            else:
                raise TypeError(f'Invalid value {value} for field {field}: expected dict')

        raise ValueError(f'Invalid field {field} for keypath {keypath}')

    @staticmethod
    def __is_set(cfg, step=None, index=None):
        '''Returns whether a user has set a value for this parameter.

        A value counts as set if a user has set a global value OR a value for
        the provided step/index.
        '''
        if Schema.GLOBAL_KEY in cfg['node'] and \
           Schema.GLOBAL_KEY in cfg['node'][Schema.GLOBAL_KEY] and \
           'value' in cfg['node'][Schema.GLOBAL_KEY][Schema.GLOBAL_KEY]:
            # global value is set
            return True

        if step is None:
            return False
        if index is None:
            index = Schema.GLOBAL_KEY

        return step in cfg['node'] and \
            index in cfg['node'][step] and \
            'value' in cfg['node'][step][index]

    @staticmethod
    def _is_leaf(cfg):
        # 'shorthelp' chosen arbitrarily: any mandatory field with a consistent
        # type would work.
        return 'shorthelp' in cfg and isinstance(cfg['shorthelp'], str)

    @staticmethod
    def __is_list(field, type):
        if field in ('filehash', 'date', 'author', 'example', 'enum', 'switch', 'package'):
            return True

        is_list = type.startswith('[')
        if is_list and field in ('signature', 'value'):
            return True

        return False

    @staticmethod
    def __validate_step_index(pernode, field, step, index):
        '''Shared validation logic for the step and index keyword arguments to
        get(), set(), and add(), based on the pernode setting of a parameter and
        field.

        Returns an error message if there's a problem with the arguments,
        otherwise None.
        '''
        if field not in Schema.PERNODE_FIELDS:
            if step is not None or index is not None:
                return 'step and index are only valid for value fields'
            return None

        if pernode == 'never' and (step is not None or index is not None):
            return 'step and index are not valid for this parameter'

        if pernode == 'required' and (step is None or index is None):
            return 'step and index are required for this parameter'

        if step is None and index is not None:
            return 'if index is provided, step must be provided as well'

        # Step and index for default should be accessed set_/get_default
        if step == 'default':
            return f'illegal step name: {step} is reserved'

        if index == 'default':
            return f'illegal index name: {step} is reserved'

        return None

    def __search(self, *keypath, insert_defaults=False, use_default=True, job=None):
        if job is not None:
            cfg = self.cfg['history'][job]
        else:
            cfg = self.cfg

        for key in keypath:
            if not isinstance(key, str):
                raise TypeError(f'Invalid keypath {keypath}: key is not a string: {key}')

            if Schema._is_leaf(cfg):
                raise ValueError(f'Invalid keypath {keypath}: unexpected key: {key}')

            if key in cfg:
                cfg = cfg[key]
            elif 'default' in cfg:
                if insert_defaults:
                    cfg[key] = copy.deepcopy(cfg['default'])
                    cfg = cfg[key]
                elif use_default:
                    cfg = cfg['default']
                else:
                    raise ValueError(f'Invalid keypath {keypath}: unexpected key: {key}')
            else:
                raise ValueError(f'Invalid keypath {keypath}: unexpected key: {key}')

        return cfg

    ###########################################################################
    def allkeys(self, *keypath_prefix):
        '''
        Returns all keypaths in the schema as a list of lists.

        See :meth:`~siliconcompiler.core.Chip.allkeys` for detailed documentation.
        '''
        if len(keypath_prefix) > 0:
            return self.__allkeys(cfg=self.getdict(*keypath_prefix))
        else:
            return self.__allkeys()

    ###########################################################################
    def __allkeys(self, cfg=None, base_key=None):
        if cfg is None:
            cfg = self.cfg

        if Schema._is_leaf(cfg):
            return []

        keylist = []
        if base_key is None:
            base_key = []
        for k in cfg:
            key = (*base_key, k)
            if Schema._is_leaf(cfg[k]):
                keylist.append(key)
            else:
                keylist.extend(self.__allkeys(cfg=cfg[k], base_key=key))
        return keylist

    ###########################################################################
    def __copyparam(self, cfgsrc, cfgdst, keypath):
        '''
        Copies a parameter into the manifest history dictionary.
        '''

        # 1. descend keypath, pop each key as its used
        # 2. create key if missing in destination dict
        # 3. populate leaf cell when keypath empty
        if keypath:
            keypath = list(keypath)
            key = keypath[0]
            keypath.pop(0)
            if key not in cfgdst.keys():
                cfgdst[key] = {}
            self.__copyparam(cfgsrc[key], cfgdst[key], keypath)
        else:
            for key in cfgsrc.keys():
                if key not in ('example', 'switch', 'help'):
                    cfgdst[key] = copy.deepcopy(cfgsrc[key])

    ###########################################################################
    def write_json(self, fout):
        localcfg = {**self.cfg}
        if self.__journal is not None:
            localcfg['__journal__'] = self.__journal
        if _has_orjson:
            manifest_str = json.dumps(localcfg, option=json.OPT_INDENT_2).decode()
        else:
            manifest_str = json.dumps(localcfg, indent=2)
        fout.write(manifest_str)

    ###########################################################################
    def write_yaml(self, fout):
        if not _has_yaml:
            raise ImportError('yaml package required to write YAML manifest')
        fout.write(yaml.dump(self.cfg, Dumper=YamlIndentDumper, default_flow_style=False))

    ###########################################################################
    def write_tcl(self, fout, prefix="", step=None, index=None, template=None):
        '''
        Prints out schema as TCL dictionary
        '''

        tcl_set_cmds = []
        for key in self.allkeys():
            typestr = self.get(*key, field='type')
            pernode = self.get(*key, field='pernode')

            if pernode == 'required' and (step is None or index is None):
                # Skip mandatory per-node parameters if step and index are not specified
                # TODO: how should we dump these?
                continue

            if pernode != 'never':
                value = self.get(*key, step=step, index=index)
            else:
                value = self.get(*key)

            # create a TCL dict
            keystr = ' '.join([escape_val_tcl(keypart, 'str') for keypart in key])

            valstr = escape_val_tcl(value, typestr)

            # Turning scalars into lists
            if not (typestr.startswith('[') or typestr.startswith('(')):
                valstr = f'[list {valstr}]'

            # TODO: Temp fix to get rid of empty args
            if valstr == '':
                valstr = '[list ]'

            outstr = f"{prefix} {keystr} {valstr}"

            # print out all non default values
            if 'default' not in key:
                tcl_set_cmds.append(outstr)

        if template:
            fout.write(template.render(manifest_dict='\n'.join(tcl_set_cmds),
                                       scroot=os.path.abspath(PACKAGE_ROOT),
                                       record_access=self._do_record_access(),
                                       record_access_id=Schema._RECORD_ACCESS_IDENTIFIER))
        else:
            for cmd in tcl_set_cmds:
                fout.write(cmd + '\n')
            fout.write('\n')

    ###########################################################################
    def write_csv(self, fout):
        if not _has_csv:
            raise RuntimeError("csv is not available")

        csvwriter = csv.writer(fout)
        csvwriter.writerow(['Keypath', 'Value'])

        allkeys = self.allkeys()
        for key in allkeys:
            keypath = ','.join(key)
            for value, step, index in self._getvals(*key):
                if step is None and index is None:
                    keypath = ','.join(key)
                elif index is None:
                    keypath = ','.join([*key, step, 'default'])
                else:
                    keypath = ','.join([*key, step, index])

                if isinstance(value, list):
                    for item in value:
                        csvwriter.writerow([keypath, item])
                else:
                    csvwriter.writerow([keypath, value])

    ###########################################################################
    def copy(self):
        '''Returns deep copy of Schema object.'''
        newscheme = Schema(cfg=self.cfg)
        if self.__journal:
            newscheme.__journal = copy.deepcopy(self.__journal)
        return newscheme

    ###########################################################################
    def prune(self):
        '''Remove all empty parameters from configuration dictionary.

        Also deletes 'help' and 'example' keys.
        '''
        # When at top of tree loop maxdepth times to make sure all stale
        # branches have been removed, not elegant, but stupid-simple
        # "good enough"

        # 10 should be enough for anyone...
        maxdepth = 10

        for _ in range(maxdepth):
            self.__prune()

    ###########################################################################
    def __prune(self, *keypath):
        '''
        Internal recursive function that creates a local copy of the Chip
        schema (cfg) with only essential non-empty parameters retained.

        '''
        cfg = self.__search(*keypath)

        # Prune when the default & value are set to the following
        # Loop through all keys starting at the top
        for k in list(cfg.keys()):
            # removing all default/template keys
            # reached a default subgraph, delete it
            if k == 'default':
                del cfg[k]
            # reached leaf-cell
            elif 'help' in cfg[k].keys():
                del cfg[k]['help']
            elif 'example' in cfg[k].keys():
                del cfg[k]['example']
            elif Schema._is_leaf(cfg[k]):
                pass
            # removing stale branches
            elif not cfg[k]:
                cfg.pop(k)
            # keep traversing tree
            else:
                self.__prune(*keypath, k)

    ###########################################################################
    def is_empty(self, *keypath):
        '''
        Utility function to check key for an empty value.
        '''
        empty = (None, [])

        values = self._getvals(*keypath)
        defvalue = self.get_default(*keypath)
        value_empty = (defvalue in empty) and \
            all([value in empty for value, _, _ in values])
        return value_empty

    ###########################################################################
    def history(self, job):
        '''
        Returns a *mutable* reference to ['history', job] as a Schema object.

        If job doesn't currently exist in history, create it with default
        values.

        Args:
            job (str): Name of historical job to return.
        '''
        if job not in self.cfg['history']:
            self.cfg['history'][job] = self._init_schema_cfg()

        # Can't initialize Schema() by passing in cfg since it performs a deep
        # copy.
        schema = Schema()
        schema.cfg = self.cfg['history'][job]
        return schema

    #######################################
    def _init_logger(self, parent=None):
        if parent:
            # If parent provided, create a child logger
            self.logger = parent.getChild('schema')
        else:
            # Check if the logger exists and create
            if not hasattr(self, 'logger') or not self.logger:
                self.logger = logging.getLogger(f'sc_schema_{id(self)}')
                self.logger.propagate = False

    #######################################
    def __getstate__(self):
        attributes = self.__dict__.copy()

        # We have to remove the chip's logger before serializing the object
        # since the logger object is not serializable.
        del attributes['logger']
        return attributes

    #######################################
    def __setstate__(self, state):
        self.__dict__ = state

        # Reinitialize logger on restore
        self._init_logger()

    #######################################
    def __record_journal(self, record_type, key, value=None, field=None, step=None, index=None):
        '''
        Record the schema transaction
        '''
        if self.__journal is None:
            return

        self.__journal.append({
            "type": record_type,
            "key": key,
            "value": value,
            "field": field,
            "step": step,
            "index": index
        })

    #######################################
    def _start_journal(self):
        '''
        Start journaling the schema transactions
        '''
        self.__journal = []

    #######################################
    def _stop_journal(self):
        '''
        Stop journaling the schema transactions
        '''
        self.__journal = None

    #######################################
    def read_journal(self, filename):
        '''
        Reads a manifest and replays the journal
        '''

        schema = Schema(logger=self.logger)
        _, schema.__journal = Schema.__read_manifest_file(str(filename))
        self._import_journal(schema)

    #######################################
    def _import_journal(self, schema):
        '''
        Import the journaled transactions from a different schema
        '''
        if not schema.__journal:
            return

        for action in schema.__journal:
            record_type = action['type']
            keypath = action['key']
            value = action['value']
            field = action['field']
            step = action['step']
            index = action['index']
            try:
                if record_type == 'set':
                    cfg = self.__search(*keypath, insert_defaults=True)
                    self.__set(*keypath, value, logger=self.logger, cfg=cfg, field=field,
                               step=step, index=index, journal_callback=None)
                elif record_type == 'add':
                    cfg = self.__search(*keypath, insert_defaults=True)
                    self._add(*keypath, value, cfg=cfg, field=field, step=step, index=index)
                elif record_type == 'unset':
                    self.unset(*keypath, step=step, index=index)
                elif record_type == 'remove':
                    self.remove(*keypath)
                else:
                    raise ValueError(f'Unknown record type {record_type}')
            except Exception as e:
                self.logger.error(f'Exception: {e}')

    #######################################
    def _do_record_access(self):
        '''
        Determine if Schema should record calls to .get
        '''
        return False

    #######################################
    def _init_record_access(self):
        '''
        Initialize record access data record
        '''
        self.__record_access = {
            "do": self._do_record_access(),
            "recording": False,
            "record": set()
        }

    #######################################
    def _start_record_access(self):
        '''
        Start recording calls to .get
        '''
        self.__record_access["recording"] = True

    #######################################
    def _stop_record_access(self):
        '''
        Stop recording calls to .get
        '''
        self.__record_access["recording"] = False

    #######################################
    def _get_record_access(self):
        '''
        Return calls to record_access
        '''
        return self.__record_access["record"].copy()

    #######################################
    def get_default(self, *keypath):
        '''Returns default value of a parameter.

        Args:
            keypath(list str): Variable length schema key list.
        '''
        cfg = self.__search(*keypath)

        if not Schema._is_leaf(cfg):
            raise ValueError(f'Invalid keypath {keypath}: get_default() '
                             'must be called on a complete keypath')

        return cfg['node']['default']['default']['value']

    #######################################
    def set_default(self, *args):
        '''Sets the default value of a parameter.

        Args:
            args (list str): Variable length schema key list and value.
        '''
        keypath = args[:-1]
        value = args[-1]
        cfg = self.__search(*keypath)

        if not Schema._is_leaf(cfg):
            raise ValueError(f'Invalid keypath {keypath}: set_default() '
                             'must be called on a complete keypath')

        allowed_values = None
        if 'enum' in cfg:
            allowed_values = cfg['enum']

        cfg['node']['default']['default']['value'] = Schema.__check_and_normalize(
            value, cfg['type'], 'value', keypath, allowed_values)

    ###########################################################################
    def create_cmdline(self,
                       progname,
                       description=None,
                       switchlist=None,
                       input_map=None,
                       additional_args=None,
                       version=None,
                       print_banner=None,
                       input_map_handler=None,
                       preprocess_keys=None,
                       post_process=None,
                       logger=None):
        """Creates a Schema command line interface.

        Exposes parameters in the SC schema as command line switches,
        simplifying creation of SC apps with a restricted set of schema
        parameters exposed at the command line. The order of command
        line switch settings parsed from the command line is as follows:

         1. loglevel, if available in schema
         2. read_manifest([cfg]), if available in schema
         3. read inputs with input_map_handler
         4. all other switches
         5. Run post_process

        The cmdline interface is implemented using the Python argparse package
        and the following use restrictions apply.

        * Help is accessed with the '-h' switch.
        * Arguments that include spaces must be enclosed with double quotes.
        * List parameters are entered individually. (ie. -y libdir1 -y libdir2)
        * For parameters with Boolean types, the switch implies "true".
        * Special characters (such as '-') must be enclosed in double quotes.
        * Compiler compatible switches include: -D, -I, -O{0,1,2,3}
        * Legacy switch formats are supported: +libext+, +incdir+

        Args:
            progname (str): Name of program to be executed.
            description (str): Short program description.
            switchlist (list of str): List of SC parameter switches to expose
                at the command line. By default all SC schema switches are
                available. Parameter switches should be entered based on the
                parameter 'switch' field in the schema. For parameters with
                multiple switches, both will be accepted if any one is included
                in this list.
            input_map (dict of str): Dictionary mapping file extensions to input
                filetypes. This is used to automatically assign positional
                source arguments to ['input', 'fileset', ...] keypaths based on their file
                extension. If None, the CLI will not accept positional source
                arguments.
            additional_args (dict of dict): Dictionary of extra arguments to add
                to the command line parser, with the arguments matching the
                argparse.add_argument() call.
            version (str): Version to report when calling with -version
            print_banner (function): Function callback to print command line banner
            input_map_handler (function): Function callback handle inputs to the input map
            preprocess_keys (function): Function callback to preprocess keys that need to be
                corrected
            post_process (function): Function callback to process arguments before returning

        Returns:
            None if additional_args is not provided, otherwise a dictionary with the
                command line options detected from the additional_args

        Examples:
            >>> schema.create_cmdline(progname='sc-show',switchlist=['-input','-cfg'])
            Creates a command line interface for 'sc-show' app.
            >>> schema.create_cmdline(progname='sc', input_map={'v': ('rtl', 'verilog')})
            All sources ending in .v will be stored in ['input', 'rtl', 'verilog']
            >>> extra = schema.create_cmdline(progname='sc',
                                              additional_args={'-demo': {'action': 'store_true'}})
            Returns extra = {'demo': False/True}
        """

        if not logger:
            logger = self.logger

        # Argparse
        parser = argparse.ArgumentParser(prog=progname,
                                         prefix_chars='-+',
                                         formatter_class=argparse.RawDescriptionHelpFormatter,
                                         description=description,
                                         allow_abbrev=False)

        # Get a new schema, in case values have already been set
        schema_class = type(self)
        schema = schema_class(logger=self.logger)

        # Iterate over all keys from an empty schema to add parser arguments
        used_switches = set()
        for keypath in schema.allkeys():
            # Fetch fields from leaf cell
            helpstr = schema.get(*keypath, field='shorthelp')
            typestr = schema.get(*keypath, field='type')
            pernodestr = schema.get(*keypath, field='pernode')

            # argparse 'dest' must be a string, so join keypath with commas
            dest = '_'.join(keypath)

            switchstrs, metavar = self.__get_switches(schema, *keypath)

            # Three switch types (bool, list, scalar)
            if not switchlist or any(switch in switchlist for switch in switchstrs):
                used_switches.update(switchstrs)
                if typestr == 'bool':
                    # Boolean type arguments
                    if pernodestr == 'never':
                        parser.add_argument(*switchstrs,
                                            nargs='?',
                                            metavar=metavar,
                                            dest=dest,
                                            const='true',
                                            help=helpstr,
                                            default=argparse.SUPPRESS)
                    else:
                        parser.add_argument(*switchstrs,
                                            metavar=metavar,
                                            nargs='?',
                                            dest=dest,
                                            action='append',
                                            help=helpstr,
                                            default=argparse.SUPPRESS)
                elif '[' in typestr or pernodestr != 'never':
                    # list type arguments
                    parser.add_argument(*switchstrs,
                                        metavar=metavar,
                                        dest=dest,
                                        action='append',
                                        help=helpstr,
                                        default=argparse.SUPPRESS)
                else:
                    # all the rest
                    parser.add_argument(*switchstrs,
                                        metavar=metavar,
                                        dest=dest,
                                        help=helpstr,
                                        default=argparse.SUPPRESS)

        print_additional_arg_value = {}
        if additional_args:
            # Add additional user specified arguments
            arg_dests = []
            for arg, arg_detail in additional_args.items():
                do_print = True
                if "sc_print" in arg_detail:
                    do_print = arg_detail["sc_print"]
                    del arg_detail["sc_print"]
                argument = parser.add_argument(arg, **arg_detail)
                print_additional_arg_value[argument.dest] = do_print

                arg_dests.append(argument.dest)
                used_switches.add(arg)
            # rewrite additional_args with new dest information
            additional_args = arg_dests

        if version:
            parser.add_argument('-version', action='version', version=version)

        # Check if there are invalid switches
        if switchlist:
            for switch in switchlist:
                if switch not in used_switches:
                    raise ValueError(f'{switch} is not a valid commandline argument')

        if input_map is not None and input_map_handler:
            parser.add_argument('source',
                                nargs='*',
                                help='Input files with filetype inferred by extension')

        # Preprocess sys.argv to enable linux commandline switch formats
        # (gcc, verilator, etc)
        scargs = []

        # Iterate from index 1, otherwise we end up with script name as a
        # 'source' positional argument
        for argument in sys.argv[1:]:
            # Split switches with one character and a number after (O0,O1,O2)
            opt = re.match(r'(\-\w)(\d+)', argument)
            # Split assign switches (-DCFG_ASIC=1)
            assign = re.search(r'(\-\w)(\w+\=\w+)', argument)
            # Split plusargs (+incdir+/path)
            plusarg = re.search(r'(\+\w+\+)(.*)', argument)
            if opt:
                scargs.append(opt.group(1))
                scargs.append(opt.group(2))
            elif plusarg:
                scargs.append(plusarg.group(1))
                scargs.append(plusarg.group(2))
            elif assign:
                scargs.append(assign.group(1))
                scargs.append(assign.group(2))
            else:
                scargs.append(argument)

        # Grab argument from pre-process sysargs
        cmdargs = vars(parser.parse_args(scargs))

        if print_banner:
            print_banner()

        extra_params = None
        if additional_args:
            # Grab user specified arguments
            extra_params = {}
            for arg in additional_args:
                if arg in cmdargs:
                    val = cmdargs[arg]
                    if print_additional_arg_value[arg] and val:
                        msg = f'Command line argument entered: "{arg}" Value: {val}'
                        self.logger.info(msg)
                    extra_params[arg] = val
                    # Remove from cmdargs
                    del cmdargs[arg]

        # Set loglevel if set at command line
        if 'option_loglevel' in cmdargs.keys():
            log_level = cmdargs['option_loglevel']
            if isinstance(log_level, list):
                # if multiple found, pick the first one
                log_level = log_level[0]
            logger.setLevel(translate_loglevel(log_level).split()[-1])

        # Read in all cfg files
        if 'option_cfg' in cmdargs.keys():
            for item in cmdargs['option_cfg']:
                self.read_manifest(item, clobber=True, clear=True, allow_missing_keys=True)

        if input_map_handler:
            # Map sources to ['input'] keypath.
            if 'source' in cmdargs:
                input_map_handler(cmdargs['source'])
                # we don't want to handle this in the next loop
                del cmdargs['source']

        # Cycle through all command args and write to manifest
        for dest, vals in cmdargs.items():
            keypath = dest.split('_')

            # Turn everything into a list for uniformity
            if not isinstance(vals, list):
                vals = [vals]

            # Cycle through all items
            for item in vals:
                if item is None:
                    # nargs=? leaves a None for booleans
                    item = ''

                if preprocess_keys:
                    item = preprocess_keys(keypath, item)

                num_free_keys = keypath.count('default')

                switches, metavar = self.__get_switches(schema, *keypath)
                switchstr = '/'.join(switches)

                if len(item.split(' ')) < num_free_keys + 1:
                    # Error out if value provided doesn't have enough words to
                    # fill in 'default' keys.
                    raise ValueError(f'Invalid value {item} for switch {switchstr}. '
                                     f'Expected format {metavar}.')

                # We replace 'default' in keypath with first N words in provided
                # value.
                *free_keys, remainder = item.split(' ', num_free_keys)
                args = [free_keys.pop(0) if key == 'default' else key for key in keypath]

                # Remainder is the value we want to set, possibly with a step/index value beforehand
                sctype = self.get(*keypath, field='type')
                pernode = self.get(*keypath, field='pernode')
                step, index = None, None
                if pernode == 'required':
                    try:
                        step, index, val = remainder.split(' ', 2)
                    except ValueError:
                        self.logger.error(f"Invalid value '{item}' for switch {switchstr}. "
                                          "Requires step and index before final value.")
                elif pernode == 'optional':
                    # Split on spaces, preserving items that are grouped in quotes
                    items = shlex.split(remainder)
                    if len(items) > 3:
                        self.logger.error(f"Invalid value '{item}'' for switch {switchstr}. "
                                          "Too many arguments, please wrap multiline "
                                          "strings in quotes.")
                        continue
                    if sctype == 'bool':
                        if len(items) == 3:
                            step, index, val = items
                        elif len(items) == 2:
                            step, val = items
                            if val != 'true' and val != 'false':
                                index = val
                                val = True
                        elif len(items) == 1:
                            val, = items
                            if val != 'true' and val != 'false':
                                step = val
                                val = True
                        else:
                            val = True
                    else:
                        if len(items) == 3:
                            step, index, val = items
                        elif len(items) == 2:
                            step, val = items
                        else:
                            val, = items
                else:
                    val = remainder

                msg = f'Command line argument entered: {args} Value: {val}'
                if step is not None:
                    msg += f' Step: {step}'
                if index is not None:
                    msg += f' Index: {index}'
                self.logger.info(msg)

                # Storing in manifest
                typestr = schema.get(*keypath, field='type')
                if typestr.startswith('['):
                    if self.valid(*args):
                        self.add(*args, val, step=step, index=index)
                    else:
                        self.set(*args, val, step=step, index=index, clobber=True)
                else:
                    self.set(*args, val, step=step, index=index, clobber=True)

        if post_process:
            extra_params = post_process(cmdargs, extra_params)

        return extra_params

    ###########################################################################
    def __get_switches(self, schema, *keypath):
        '''Helper function for parsing switches and metavars for a keypath.'''
        # Switch field fully describes switch format
        switch = schema.get(*keypath, field='switch')

        if switch is None:
            switches = []
        elif isinstance(switch, list):
            switches = switch
        else:
            switches = [switch]
        switchstrs = []

        # parse out switch from metavar
        # TODO: should we validate that metavar matches for each switch?
        for switch in switches:
            switchmatch = re.match(r'(-[\w_]+)\s+(.*)', switch)
            gccmatch = re.match(r'(-[\w_]+)(.*)', switch)
            plusmatch = re.match(r'(\+[\w_\+]+)(.*)', switch)

            if switchmatch:
                switchstr = switchmatch.group(1)
                metavar = switchmatch.group(2)
            elif gccmatch:
                switchstr = gccmatch.group(1)
                metavar = gccmatch.group(2)
            elif plusmatch:
                switchstr = plusmatch.group(1)
                metavar = plusmatch.group(2)
            switchstrs.append(switchstr)

        return switchstrs, metavar

    ###########################################################################
    def read_manifest(self, filename, clear=True, clobber=True, allow_missing_keys=True):
        """
        Reads a manifest from disk and merges it with the current manifest.

        The file format read is determined by the filename suffix. Currently
        json (*.json) and yaml(*.yaml) formats are supported.

        Args:
            filename (filepath): Path to a manifest file to be loaded.
            clear (bool): If True, disables append operations for list type.
            clobber (bool): If True, overwrites existing parameter value.
            allow_missing_keys (bool): If True, keys not present in current schema will be ignored.

        Examples:
            >>> chip.read_manifest('mychip.json')
            Loads the file mychip.json into the current Chip object.
        """
        schema = Schema(manifest=filename, logger=self.logger)

        if schema.get('schemaversion') != self.get('schemaversion'):
            self.logger.warning("Mismatch in schema versions: "
                                f"{schema.get('schemaversion')} != {self.get('schemaversion')}")

        for keylist in schema.allkeys():
            if keylist[0] in ('history', 'library'):
                continue
            if 'default' in keylist:
                continue
            typestr = schema.get(*keylist, field='type')
            should_append = '[' in typestr and not clear

            if allow_missing_keys and not self.valid(*keylist, default_valid=True):
                self.logger.warning(f'{keylist} not found in schema, skipping...')
                continue

            for val, step, index in schema._getvals(*keylist, return_defvalue=False):
                # update value, handling scalars vs. lists
                if should_append:
                    self.add(*keylist, val, step=step, index=index)
                else:
                    self.set(*keylist, val, step=step, index=index, clobber=clobber)

                # update other pernode fields
                # TODO: only update these if clobber is successful
                step_key = Schema.GLOBAL_KEY if not step else step
                idx_key = Schema.GLOBAL_KEY if not index else index
                for field in schema.getdict(*keylist)['node'][step_key][idx_key].keys():
                    if field == 'value':
                        continue
                    v = schema.get(*keylist, step=step, index=index, field=field)
                    if should_append:
                        self.add(*keylist, v, step=step, index=index, field=field)
                    else:
                        self.set(*keylist, v, step=step, index=index, field=field)

            # update other fields that a user might modify
            for field in schema.getdict(*keylist).keys():
                if field in ('node',):
                    # skip these fields (node handled above)
                    continue

                # TODO: should we be taking into consideration clobber for these fields?
                v = schema.get(*keylist, field=field)
                self.set(*keylist, v, field=field)

        # Read history, if we're not already reading into a job
        if 'history' in schema.getkeys():
            for historic_job in schema.getkeys('history'):
                self.cfg['history'][historic_job] = schema.getdict('history', historic_job)

        # TODO: better way to handle this?
        if 'library' in schema.getkeys():
            for libname in schema.getkeys('library'):
                self.cfg['library'][libname] = schema.getdict('library', libname)

    ###########################################################################
    def merge_manifest(self, src, job=None, clobber=True, clear=True, check=False):
        """
        Merges a given manifest with the current compilation manifest.

        All value fields in the provided schema dictionary are merged into the
        current chip object. Dictionaries with non-existent keypath produces a
        logger error message and raises the Chip object error flag.

        Args:
            src (Schema): Schema object to merge
            job (str): Specifies non-default job to merge into
            clear (bool): If True, disables append operations for list type
            clobber (bool): If True, overwrites existing parameter value
            check (bool): If True, checks the validity of each key
        """
        if job is not None:
            dest = self.history(job)
        else:
            dest = self

        for keylist in src.allkeys():
            if keylist[0] in ('history', 'library'):
                continue
            # only read in valid keypaths without 'default'
            key_valid = True
            if check:
                key_valid = dest.valid(*keylist, default_valid=True)
                if not key_valid:
                    self.logger.warning(f'Keypath {keylist} is not valid')
            if key_valid and 'default' not in keylist:
                typestr = src.get(*keylist, field='type')
                should_append = '[' in typestr and not clear
                key_cfg = src.__search(*keylist)
                for val, step, index in src._getvals(*keylist, return_defvalue=False):
                    # update value, handling scalars vs. lists
                    if should_append:
                        dest.add(*keylist, val, step=step, index=index)
                    else:
                        dest.set(*keylist, val, step=step, index=index, clobber=clobber)

                    # update other pernode fields
                    # TODO: only update these if clobber is successful
                    step_key = Schema.GLOBAL_KEY if not step else step
                    idx_key = Schema.GLOBAL_KEY if not index else index
                    for field in key_cfg['node'][step_key][idx_key].keys():
                        if field == 'value':
                            continue
                        v = src.get(*keylist, step=step, index=index, field=field)
                        if should_append:
                            dest.add(*keylist, v, step=step, index=index, field=field)
                        else:
                            dest.set(*keylist, v, step=step, index=index, field=field)

                # update other fields that a user might modify
                for field in key_cfg.keys():
                    if field in ('node', 'switch', 'type', 'require',
                                 'shorthelp', 'example', 'help'):
                        # skip these fields (node handled above, others are static)
                        continue
                    # TODO: should we be taking into consideration clobber for these fields?
                    v = src.get(*keylist, field=field)
                    dest.set(*keylist, v, field=field)


if _has_yaml:
    class YamlIndentDumper(yaml.Dumper):
        def increase_indent(self, flow=False, indentless=False):
            return super(YamlIndentDumper, self).increase_indent(flow, False)



'''
Xyce XDM netlist converter, used to convert PSPICE and HSPICE netists into Xyce format.

Sources: https://github.com/Xyce/XDM

Installation: https://github.com/Xyce/XDM

Status: SC integration WIP
'''
from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'exe', 'xdm_bdl')
    chip.set('tool', tool, 'vswitch', '-h')
    chip.set('tool', tool, 'version', '>=v2.7.0', clobber=False)


################################
# Version Check
################################
def parse_version(stdout):
    line = stdout.splitlines()[5]
    return line.split()[1]



from siliconcompiler.tools.surelog import parse as surelog_parse
from siliconcompiler.tools.chisel import convert as chisel_convert
from siliconcompiler.tools.bambu import convert as bambu_convert
from siliconcompiler.tools.bluespec import convert as bluespec_convert
from siliconcompiler.tools.ghdl import convert as ghdl_convert
from siliconcompiler.tools.sv2v import convert as sv2v_convert

from siliconcompiler.tools.builtin import concatenate


def _make_docs(chip):
    from siliconcompiler.targets import freepdk45_demo
    chip.set('input', 'rtl', 'vhdl', 'test')
    chip.set('input', 'rtl', 'verilog', 'test')
    chip.set('input', 'rtl', 'systemverilog', 'test')
    chip.set('input', 'hll', 'c', 'test')
    chip.set('input', 'hll', 'bsv', 'test')
    chip.set('input', 'hll', 'scala', 'test')

    chip.use(freepdk45_demo)


def __get_frontends(allow_system_verilog):
    systemverilog_frontend = [
        ('import', surelog_parse)
    ]
    if not allow_system_verilog:
        systemverilog_frontend.append(('convert', sv2v_convert))

    return {
        "verilog": systemverilog_frontend,
        "chisel": [('import', chisel_convert)],
        "c": [('import', bambu_convert)],
        "bluespec": [('import', bluespec_convert)],
        "vhdl": [('import', ghdl_convert)]
    }


def setup_multiple_frontends(flow, allow_system_verilog=False):
    '''
    Sets of multiple frontends if different frontends are required.

    Returns name of final step from the setup.
    '''

    concat_nodes = []
    flowname = flow.design
    for frontend, pipe in __get_frontends(allow_system_verilog).items():
        prev_step = None
        for step, task in pipe:
            step_name = f'{step}_{frontend}'

            flow.node(flowname, step_name, task)
            if prev_step:
                flow.edge(flowname, prev_step, step_name)

            prev_step = step_name

        if prev_step:
            concat_nodes.append(prev_step)

    final_node = 'combine'
    flow.node(flowname, final_node, concatenate)
    for node in concat_nodes:
        flow.edge(flowname, node, final_node)

    return final_node
